"""Settings used for multiple tests."""

# Default tolerances
TOLERANCES = dict(atol=1e-3, rtol=1e-4)
MILD_TOLERANCES = dict(atol=0.05, rtol=0.05)
STRICT_TOLERANCES = dict(atol=1e-6, rtol=1e-6)

# Batch dimensions that are typically checked
BATCH_DIMS = [[7, 9]]



import pytest
import torch

from lgatr.layers.mlp.nonlinearities import ScalarGatedNonlinearity
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("num_scalars", [9])
@pytest.mark.parametrize("activation", ["gelu", "relu", "sigmoid", "silu"])
def test_scalar_gated_nonlinearity_layer_equivariance(activation, batch_dims, num_scalars):
    """Tests ScalarGatedNonlinearity() for equivariance."""
    layer = ScalarGatedNonlinearity(nonlinearity=activation)
    scalars = torch.randn(*batch_dims, num_scalars)
    check_pin_equivariance(
        layer, 1, batch_dims=batch_dims, fn_kwargs=dict(scalars=scalars), **TOLERANCES
    )



import pytest
import torch

from lgatr.layers.attention.config import SelfAttentionConfig
from lgatr.layers.mlp.config import MLPConfig
from lgatr.nets import LGATr
from lgatr.primitives.config import gatr_config
from tests.helpers import BATCH_DIMS, MILD_TOLERANCES, check_pin_equivariance

S_CHANNELS = [(None, None, 7), (4, 5, 6)]


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize(
    "num_items,in_mv_channels,out_mv_channels,hidden_mv_channels", [(8, 3, 4, 6)]
)
@pytest.mark.parametrize("num_heads,num_blocks", [(4, 1)])
@pytest.mark.parametrize("in_s_channels,out_s_channels,hidden_s_channels", S_CHANNELS)
@pytest.mark.parametrize("dropout_prob", [None, 0.0, 0.3])
@pytest.mark.parametrize("multi_query_attention", [False, True])
@pytest.mark.parametrize("checkpoint_blocks", [False, True])
@pytest.mark.parametrize("use_fully_connected_subgroup", [True, False])
def test_lgatr_shape(
    batch_dims,
    num_items,
    in_mv_channels,
    out_mv_channels,
    hidden_mv_channels,
    num_blocks,
    num_heads,
    in_s_channels,
    out_s_channels,
    hidden_s_channels,
    multi_query_attention,
    dropout_prob,
    checkpoint_blocks,
    use_fully_connected_subgroup,
):
    """Tests the output shape of LGATr."""
    gatr_config.use_fully_connected_subgroup = use_fully_connected_subgroup

    inputs = torch.randn(*batch_dims, num_items, in_mv_channels, 16)
    scalars = None if in_s_channels is None else torch.randn(*batch_dims, num_items, in_s_channels)

    try:
        net = LGATr(
            in_mv_channels=in_mv_channels,
            out_mv_channels=out_mv_channels,
            hidden_mv_channels=hidden_mv_channels,
            in_s_channels=in_s_channels,
            out_s_channels=out_s_channels,
            hidden_s_channels=hidden_s_channels,
            attention=dict(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            num_blocks=num_blocks,
            mlp=dict(),
            dropout_prob=dropout_prob,
            checkpoint_blocks=checkpoint_blocks,
        )
    except NotImplementedError:
        # Some features require scalar inputs, and failing without them is fine
        return

    outputs, output_scalars = net(inputs, scalars=scalars)

    assert outputs.shape == (*batch_dims, num_items, out_mv_channels, 16)
    if in_s_channels is not None:
        assert output_scalars.shape == (*batch_dims, num_items, out_s_channels)

    # restore defaults
    gatr_config.use_fully_connected_subgroup = True


@pytest.mark.parametrize("batch_dims", [(64,)])
@pytest.mark.parametrize(
    "num_items,in_mv_channels,out_mv_channels,hidden_mv_channels", [(8, 3, 4, 6)]
)
@pytest.mark.parametrize("num_heads,num_blocks", [(4, 1)])
@pytest.mark.parametrize("in_s_channels,out_s_channels,hidden_s_channels", S_CHANNELS)
@pytest.mark.parametrize("multi_query_attention", [False, True])
def test_lgatr_equivariance(
    batch_dims,
    num_items,
    in_mv_channels,
    out_mv_channels,
    hidden_mv_channels,
    num_blocks,
    num_heads,
    in_s_channels,
    out_s_channels,
    hidden_s_channels,
    multi_query_attention,
):
    """Tests LGATr for equivariance."""
    try:
        net = LGATr(
            in_mv_channels=in_mv_channels,
            out_mv_channels=out_mv_channels,
            hidden_mv_channels=hidden_mv_channels,
            in_s_channels=in_s_channels,
            out_s_channels=out_s_channels,
            hidden_s_channels=hidden_s_channels,
            attention=SelfAttentionConfig(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            num_blocks=num_blocks,
            mlp=MLPConfig(),
        )
    except NotImplementedError:
        # Some features require scalar inputs, and failing without them is fine
        return

    scalars = None if in_s_channels is None else torch.randn(*batch_dims, num_items, in_s_channels)
    data_dims = tuple(list(batch_dims) + [num_items, in_mv_channels])
    check_pin_equivariance(
        net, 1, batch_dims=data_dims, fn_kwargs=dict(scalars=scalars), **MILD_TOLERANCES
    )



import pytest
import torch

from lgatr.nets.conditional_lgatr_slim import (
    ConditionalLGATrSlim,
    ConditionalLGATrSlimBlock,
    CrossAttention,
)

from ...helpers.constants import BATCH_DIMS, TOLERANCES
from ...helpers.equivariance_noga import check_equivariance

BATCH_DIMS = [b[:-1] for b in BATCH_DIMS]


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("N,Ncond", [(3, 7), (13, 2)])
@pytest.mark.parametrize("v_channels,cond_v_channels,s_channels,cond_s_channels", [(24, 6, 14, 20)])
@pytest.mark.parametrize("num_heads,attn_ratio", [(2, 1), (1, 2)])
def test_CrossAttention_equivariance(
    batch_dims,
    N,
    Ncond,
    v_channels,
    cond_v_channels,
    s_channels,
    cond_s_channels,
    num_heads,
    attn_ratio,
):
    layer = CrossAttention(
        q_v_channels=v_channels,
        kv_v_channels=cond_v_channels,
        q_s_channels=s_channels,
        kv_s_channels=cond_s_channels,
        num_heads=num_heads,
        attn_ratio=attn_ratio,
    )
    s = torch.randn(*batch_dims, N, s_channels)
    cond_s = torch.randn(*batch_dims, Ncond, cond_s_channels)

    v = torch.randn(*batch_dims, N, v_channels, 4)
    cond_v = torch.randn(*batch_dims, Ncond, cond_v_channels, 4)
    out_v, out_s = layer(v, cond_v, s, cond_s)
    assert out_v.shape == v.shape
    assert out_s.shape == s.shape

    batch_dims = [batch_dims + [N, v_channels], batch_dims + [Ncond, cond_v_channels]]
    check_equivariance(
        layer,
        batch_dims=batch_dims,
        num_args=2,
        fn_kwargs=dict(scalars=s, scalars_condition=cond_s),
        **TOLERANCES,
    )


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("N,Ncond", [(3, 7), (13, 2)])
@pytest.mark.parametrize(
    "v_channels,cond_v_channels,s_channels,cond_s_channels,num_heads", [(24, 6, 14, 20, 2)]
)
@pytest.mark.parametrize("dropout_prob", [None, 0.0, 0.5])
def test_ConditionalLGATrSlimBlock_equivariance(
    batch_dims,
    N,
    Ncond,
    v_channels,
    cond_v_channels,
    s_channels,
    cond_s_channels,
    num_heads,
    dropout_prob,
):
    layer = ConditionalLGATrSlimBlock(
        v_channels=v_channels,
        condition_v_channels=cond_v_channels,
        s_channels=s_channels,
        condition_s_channels=cond_s_channels,
        num_heads=num_heads,
        dropout_prob=dropout_prob,
    )
    layer.eval()

    s = torch.randn(*batch_dims, N, s_channels)
    cond_s = torch.randn(*batch_dims, Ncond, cond_s_channels)
    batch_dims = [batch_dims + [N, v_channels], batch_dims + [Ncond, cond_v_channels]]

    # equivariance
    check_equivariance(
        layer,
        batch_dims=batch_dims,
        num_args=2,
        fn_kwargs=dict(scalars=s, scalars_condition=cond_s),
        **TOLERANCES,
    )


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("N,Ncond", [(3, 7), (13, 2)])
@pytest.mark.parametrize(
    "in_v_channels,in_s_channels,out_v_channels,out_s_channels,cond_v_channels,cond_s_channels",
    [
        (4, 3, 9, 2, 5, 6),
        (2, 9, 0, 3, 4, 7),
        (3, 5, 7, 0, 6, 8),
        (8, 3, 0, 0, 7, 9),
    ],
)
@pytest.mark.parametrize("hidden_v_channels,hidden_s_channels,num_heads", [(32, 4, 1), (16, 8, 4)])
@pytest.mark.parametrize("dropout_prob", [None, 0.0, 0.5])
@pytest.mark.parametrize("num_blocks", [1, 2])
@pytest.mark.parametrize("checkpoint_blocks", [False, True])
def test_ConditionalLGATrSlim_equivariance(
    batch_dims,
    N,
    Ncond,
    in_v_channels,
    in_s_channels,
    out_v_channels,
    out_s_channels,
    cond_v_channels,
    cond_s_channels,
    hidden_v_channels,
    hidden_s_channels,
    num_heads,
    num_blocks,
    dropout_prob,
    checkpoint_blocks,
):
    layer = ConditionalLGATrSlim(
        in_v_channels=in_v_channels,
        condition_v_channels=cond_v_channels,
        out_v_channels=out_v_channels,
        hidden_v_channels=hidden_v_channels,
        in_s_channels=in_s_channels,
        condition_s_channels=cond_s_channels,
        out_s_channels=out_s_channels,
        hidden_s_channels=hidden_s_channels,
        num_blocks=num_blocks,
        num_heads=num_heads,
        dropout_prob=dropout_prob,
        checkpoint_blocks=checkpoint_blocks,
    )
    layer.eval()

    s = torch.randn(*batch_dims, N, in_s_channels)
    cond_s = torch.randn(*batch_dims, Ncond, cond_s_channels)
    v = torch.randn(*batch_dims, N, in_v_channels, 4)
    cond_v = torch.randn(*batch_dims, Ncond, cond_v_channels, 4)

    out_v, out_s = layer(
        vectors=v,
        vectors_condition=cond_v,
        scalars=s,
        scalars_condition=cond_s,
    )
    assert out_v.shape == v.shape[:-2] + (out_v_channels, 4)
    assert out_s.shape == s.shape[:-1] + (out_s_channels,)

    # equivariance
    batch_dims = [batch_dims + [N, in_v_channels], batch_dims + [Ncond, cond_v_channels]]
    check_equivariance(
        layer,
        batch_dims=batch_dims,
        num_args=2,
        fn_kwargs=dict(scalars=s, scalars_condition=cond_s),
        **TOLERANCES,
    )


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("N,Ncond", [(3, 7)])
@pytest.mark.parametrize(
    "in_v_channels,in_s_channels,out_v_channels,out_s_channels,cond_v_channels,cond_s_channels",
    [(4, 3, 9, 2, 5, 6)],
)
@pytest.mark.parametrize(
    "hidden_v_channels,hidden_s_channels,num_heads,num_blocks", [(16, 8, 4, 1)]
)
def test_ConditionalLGATrSlim_equivariance_compiled(
    batch_dims,
    N,
    Ncond,
    in_v_channels,
    in_s_channels,
    out_v_channels,
    out_s_channels,
    cond_v_channels,
    cond_s_channels,
    hidden_v_channels,
    hidden_s_channels,
    num_heads,
    num_blocks,
    compile=True,
):
    layer = ConditionalLGATrSlim(
        in_v_channels=in_v_channels,
        condition_v_channels=cond_v_channels,
        out_v_channels=out_v_channels,
        hidden_v_channels=hidden_v_channels,
        in_s_channels=in_s_channels,
        condition_s_channels=cond_s_channels,
        out_s_channels=out_s_channels,
        hidden_s_channels=hidden_s_channels,
        num_blocks=num_blocks,
        num_heads=num_heads,
        compile=compile,
    )
    layer.eval()

    s = torch.randn(*batch_dims, N, in_s_channels)
    cond_s = torch.randn(*batch_dims, Ncond, cond_s_channels)
    v = torch.randn(*batch_dims, N, in_v_channels, 4)
    cond_v = torch.randn(*batch_dims, Ncond, cond_v_channels, 4)

    out_v, out_s = layer(
        vectors=v,
        vectors_condition=cond_v,
        scalars=s,
        scalars_condition=cond_s,
    )
    assert out_v.shape == v.shape[:-2] + (out_v_channels, 4)
    assert out_s.shape == s.shape[:-1] + (out_s_channels,)

    # equivariance
    batch_dims = [batch_dims + [N, in_v_channels], batch_dims + [Ncond, cond_v_channels]]
    check_equivariance(
        layer,
        batch_dims=batch_dims,
        num_args=2,
        fn_kwargs=dict(scalars=s, scalars_condition=cond_s),
        **TOLERANCES,
    )



import pytest
import torch

from lgatr.interface import embed_pseudoscalar, extract_pseudoscalar
from tests.helpers import BATCH_DIMS, TOLERANCES


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_embed_pseudoscalar(batch_dims):
    """Tests that embed_pseudoscalar() embeds pseudoscalars into multivectors correctly."""

    pseudoscalar = torch.randn(*batch_dims, 1)
    mv = embed_pseudoscalar(pseudoscalar)

    # Check that scalar part of MV contains the data we wanted to embed
    mv_pseudoscalar = mv[..., [15]]
    torch.testing.assert_close(mv_pseudoscalar, pseudoscalar, **TOLERANCES)

    # Check that other components of MV are empty
    mv_other = mv[..., :-1]
    torch.testing.assert_close(mv_other, torch.zeros_like(mv_other), **TOLERANCES)

    # Check that we can extract the pseudoscalar back from the multivector
    other_pseudoscalar = extract_pseudoscalar(embed_pseudoscalar(pseudoscalar))
    torch.testing.assert_close(pseudoscalar, other_pseudoscalar, **TOLERANCES)



import pytest
import torch

from lgatr.layers.layer_norm import EquiLayerNorm
from lgatr.primitives import abs_squared_norm
from tests.helpers import TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", [(7, 9)])
@pytest.mark.parametrize("num_scalars", [9])
def test_equi_layer_norm_layer_correctness(batch_dims, num_scalars):
    """Tests whether the output of EquiLayerNorm has the correct variance."""
    inputs = torch.randn(*batch_dims, 16)
    scalars = torch.randn(*batch_dims, num_scalars)
    layer = EquiLayerNorm(epsilon=1e-9)
    normalized_inputs, _ = layer(inputs, scalars=scalars)
    dims = tuple(range(1, len(batch_dims) + 1))
    variance = torch.mean(abs_squared_norm(normalized_inputs), dim=dims)
    torch.testing.assert_close(variance, torch.ones_like(variance), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", [(7, 9)])
@pytest.mark.parametrize("num_scalars", [9])
def test_equi_layer_norm_layer_equivariance(batch_dims, num_scalars):
    """Tests EquiLayerNorm() for equivariance."""
    layer = EquiLayerNorm()
    scalars = torch.randn(*batch_dims, num_scalars)
    check_pin_equivariance(
        layer, 1, batch_dims=batch_dims, fn_kwargs=dict(scalars=scalars), **TOLERANCES
    )



"""Equivariant transformer for vector and scalar data."""

import math

import torch
from torch import nn
from torch.nn.functional import dropout, dropout1d
from torch.utils.checkpoint import checkpoint

from ..primitives.attention import scaled_dot_product_attention
from ..utils.misc import minimum_autocast_precision


def inner_product(x, y):
    t = x[..., 0] * y[..., 0]
    s = (x[..., 1:] * y[..., 1:]).sum(dim=-1)
    return t - s


def squared_norm(x):
    return inner_product(x, x)


def get_nonlinearity(label):
    if label == "relu":
        return nn.ReLU()
    elif label == "sigmoid":
        return nn.Sigmoid()
    elif label == "tanh":
        return nn.Tanh()
    elif label == "gelu":
        return nn.GELU()
    elif label == "silu":
        return nn.SiLU()
    else:
        raise ValueError(f"Unsupported nonlinearity type: {label}")


class Dropout(nn.Module):
    """Dropout module for scalar and vector features.

    For vector features, the same dropout mask is applied to all four components of each vector.
    """

    def __init__(self, dropout_prob: float):
        super().__init__()
        self._dropout_prob = dropout_prob

    def forward(self, vectors, scalars):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the dropped out vectors and scalars.
        """
        # have to reshape vectors because dropout1d constrains input shape
        v = vectors.reshape(-1, 4)
        out_v = dropout1d(v, p=self._dropout_prob, training=self.training)
        out_v = out_v.reshape(vectors.shape)

        out_s = dropout(scalars, p=self._dropout_prob, training=self.training)
        return out_v, out_s


class RMSNorm(nn.Module):
    """Normalize jointly over vector and scalar features.

    For vectors, we use the absolute value of the squared norm because otherwise negative norms are possible.
    """

    def __init__(self, epsilon: float = 0.01):
        super().__init__()
        self.epsilon = epsilon

    @minimum_autocast_precision(torch.float32)
    def forward(self, vectors, scalars):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        v_squared_norm = squared_norm(vectors).abs()
        s_squared_norm = scalars.square()
        sum_squared_norms = v_squared_norm.sum(dim=-1) + s_squared_norm.sum(dim=-1)
        mean_squared_norms = sum_squared_norms / (vectors.shape[-2] + scalars.shape[-1])
        norm = torch.rsqrt(mean_squared_norms + self.epsilon).unsqueeze(-1)

        vectors_out = vectors * norm.unsqueeze(-1)
        scalars_out = scalars * norm
        return vectors_out, scalars_out


class Linear(nn.Module):
    """Linear operations for vector and scalar features.

    Supports optional mixing between vector and scalar features to improve expressivity.
    """

    def __init__(
        self,
        in_v_channels: int,
        out_v_channels: int,
        in_s_channels: int,
        out_s_channels: int,
        bias: bool = True,
        initialization: str = "default",
    ):
        """
        Parameters
        ----------
        in_v_channels : int
            Number of input vector channels.
        out_v_channels : int
            Number of output vector channels.
        in_s_channels : int
            Number of input scalar channels.
        out_s_channels : int
            Number of output scalar channels.
        bias : bool, optional
            Whether to include a bias term in the scalar linear layer, by default True.
        initialization : str, optional
            Initialization method for weights, by default "default".
            The alternative "small" initializes weights to smaller values,
            which might improve stability in attention projections.
        """
        super().__init__()
        self._in_v_channels = in_v_channels
        self._out_v_channels = out_v_channels
        self._in_s_channels = in_s_channels
        self._out_s_channels = out_s_channels
        self._bias = bias

        self.weight_v = nn.Parameter(
            torch.empty(
                (
                    out_v_channels,
                    in_v_channels,
                )
            )
        )
        self.linear_s = nn.Linear(in_s_channels, out_s_channels, bias=bias)

        self.reset_parameters(initialization)

    def forward(self, vectors, scalars):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        vectors_out = self.weight_v @ vectors
        scalars_out = self.linear_s(scalars)
        return vectors_out, scalars_out

    def reset_parameters(self, initialization, additional_factor=1.0):
        if initialization == "default":
            v_factor = additional_factor
            s_factor = additional_factor
        elif initialization == "small":
            v_factor = 0.1 * additional_factor
            s_factor = 0.1 * additional_factor
        else:
            raise ValueError(f"Unknown initialization: {initialization}")

        fan_in = max(self._in_v_channels, 1)
        bound = v_factor / math.sqrt(fan_in)
        nn.init.uniform_(self.weight_v, a=-bound, b=bound)

        fan_in = max(self._in_s_channels, 1)
        bound = s_factor / math.sqrt(fan_in)
        nn.init.uniform_(self.linear_s.weight, a=-bound, b=bound)


class GatedLinearUnit(nn.Module):
    """Gated linear unit (GLU) for vector and scalar features.

    Scalar gates are computed from scalar features,
    while vector gates are computed from inner products of vector features.
    """

    def __init__(
        self,
        in_v_channels: int,
        out_v_channels: int,
        in_s_channels: int,
        out_s_channels: int,
        nonlinearity: str = "gelu",
    ):
        super().__init__()
        self.linear = Linear(
            in_v_channels=in_v_channels,
            out_v_channels=3 * out_v_channels,
            in_s_channels=in_s_channels,
            out_s_channels=2 * out_s_channels,
        )
        self.nonlinearity = get_nonlinearity(nonlinearity)

    def forward(self, vectors, scalars):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        v_full, s_full = self.linear(vectors, scalars)
        v_pre, v_gates_1, v_gates_2 = v_full.chunk(3, dim=-2)
        s_pre, s_gates = s_full.chunk(2, dim=-1)

        v_gates = inner_product(v_gates_1, v_gates_2).unsqueeze(-1)
        vectors_out = self.nonlinearity(v_gates) * v_pre
        scalars_out = self.nonlinearity(s_gates) * s_pre
        return vectors_out, scalars_out


class SelfAttention(nn.Module):
    """Self-attention module for Lorentz vectors and scalar features."""

    def __init__(
        self,
        v_channels: int,
        s_channels: int,
        num_heads: int,
        attn_ratio: int = 1,
        dropout_prob: float | None = None,
    ):
        super().__init__()
        self.hidden_v_channels = max(attn_ratio * v_channels // num_heads, 1)
        self.hidden_s_channels = max(attn_ratio * s_channels // num_heads, 4)
        self.num_heads = num_heads

        metric = torch.tensor([1.0, -1.0, -1.0, -1.0])
        self.register_buffer("metric", metric)

        self.linear_in = Linear(
            in_v_channels=v_channels,
            out_v_channels=3 * self.hidden_v_channels * self.num_heads,
            in_s_channels=s_channels,
            out_s_channels=3 * self.hidden_s_channels * self.num_heads,
            initialization="small",
        )
        self.linear_out = Linear(
            in_v_channels=self.hidden_v_channels * self.num_heads,
            out_v_channels=v_channels,
            in_s_channels=self.hidden_s_channels * self.num_heads,
            out_s_channels=s_channels,
            initialization="small",
        )
        self.norm = RMSNorm()
        if dropout_prob is not None:
            self.dropout = Dropout(dropout_prob)
        else:
            self.dropout = None

    def _pre_reshape(self, qkv_v, qkv_s):
        qkv_v = (
            qkv_v.unflatten(-2, (3, self.hidden_v_channels, self.num_heads))
            .movedim(-4, 0)
            .movedim(-2, -4)
        )
        qkv_s = (
            qkv_s.unflatten(-1, (3, self.hidden_s_channels, self.num_heads))
            .movedim(-3, 0)
            .movedim(-1, -3)
        )

        # normalize for stability (important)
        qkv_v, qkv_s = self.norm(qkv_v, qkv_s)

        q_v, k_v, v_v = qkv_v.unbind(0)
        q_s, k_s, v_s = qkv_s.unbind(0)

        q_v_mod = q_v * self.metric
        q = torch.cat([q_v_mod.flatten(start_dim=-2), q_s], dim=-1)
        k = torch.cat([k_v.flatten(start_dim=-2), k_s], dim=-1)
        v = torch.cat([v_v.flatten(start_dim=-2), v_s], dim=-1)
        return q, k, v

    def _post_reshape(self, out):
        h_v = out[..., : self.hidden_v_channels * 4].reshape(
            *out.shape[:-1], self.hidden_v_channels, 4
        )
        h_s = out[..., self.hidden_v_channels * 4 :]

        h_v = h_v.movedim(-3, -4).flatten(-3, -2)
        h_s = h_s.movedim(-2, -3).flatten(-2, -1)
        return h_v, h_s

    def forward(self, vectors, scalars, **attn_kwargs):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.
        **attn_kwargs : dict
            Additional keyword arguments for the attention function.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        qkv_v, qkv_s = self.linear_in(vectors, scalars)

        q, k, v = self._pre_reshape(qkv_v, qkv_s)
        out = scaled_dot_product_attention(q, k, v, **attn_kwargs)
        h_v, h_s = self._post_reshape(out)

        out_v, out_s = self.linear_out(h_v, h_s)

        if self.dropout is not None:
            out_v, out_s = self.dropout(out_v, out_s)
        return out_v, out_s


class MLP(nn.Module):
    """Multi-layer perceptron (MLP) for vector and scalar features."""

    def __init__(
        self,
        v_channels: int,
        s_channels: int,
        nonlinearity: str = "gelu",
        mlp_ratio: int = 2,
        num_layers: int = 2,
        dropout_prob: float | None = None,
    ):
        super().__init__()
        assert num_layers >= 2
        layers = []

        v_channels_list = [v_channels] + [mlp_ratio * v_channels] * (num_layers - 1) + [v_channels]
        s_channels_list = [s_channels] + [mlp_ratio * s_channels] * (num_layers - 1) + [s_channels]

        for i in range(num_layers - 1):
            layers.append(
                GatedLinearUnit(
                    in_v_channels=v_channels_list[i],
                    out_v_channels=v_channels_list[i + 1],
                    in_s_channels=s_channels_list[i],
                    out_s_channels=s_channels_list[i + 1],
                    nonlinearity=nonlinearity,
                )
            )
            if dropout_prob is not None:
                layers.append(Dropout(dropout_prob))
        layers.append(
            Linear(
                in_v_channels=v_channels_list[-2],
                out_v_channels=v_channels_list[-1],
                in_s_channels=s_channels_list[-2],
                out_s_channels=s_channels_list[-1],
            )
        )

        self.layers = nn.ModuleList(layers)

    def forward(self, vectors, scalars):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        v, s = vectors, scalars

        for layer in self.layers:
            v, s = layer(v, scalars=s)

        return v, s


class LGATrSlimBlock(nn.Module):
    """A single block of the L-GATr-slim,
    consisting of self-attention and MLP layers, pre-norm and residual connections."""

    def __init__(
        self,
        v_channels: int,
        s_channels: int,
        num_heads: int,
        nonlinearity: str = "gelu",
        mlp_ratio: int = 2,
        attn_ratio: int = 1,
        num_layers_mlp: int = 2,
        dropout_prob: float | None = None,
    ):
        super().__init__()

        self.norm = RMSNorm()

        self.attention = SelfAttention(
            v_channels=v_channels,
            s_channels=s_channels,
            num_heads=num_heads,
            attn_ratio=attn_ratio,
            dropout_prob=dropout_prob,
        )

        self.mlp = MLP(
            v_channels=v_channels,
            s_channels=s_channels,
            nonlinearity=nonlinearity,
            mlp_ratio=mlp_ratio,
            num_layers=num_layers_mlp,
            dropout_prob=dropout_prob,
        )

    def forward(self, vectors, scalars, **attn_kwargs):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.
        **attn_kwargs : dict
            Additional keyword arguments for the attention function.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        h_v, h_s = self.norm(vectors, scalars)

        h_v, h_s = self.attention(
            h_v,
            h_s,
            **attn_kwargs,
        )

        outputs_v = vectors + h_v
        outputs_s = scalars + h_s

        h_v, h_s = self.norm(outputs_v, outputs_s)

        h_v, h_s = self.mlp(h_v, h_s)

        outputs_v = outputs_v + h_v
        outputs_s = outputs_s + h_s

        return outputs_v, outputs_s


class LGATrSlim(nn.Module):
    """L-GATr-slim network."""

    def __init__(
        self,
        in_v_channels: int,
        out_v_channels: int,
        hidden_v_channels: int,
        in_s_channels: int,
        out_s_channels: int,
        hidden_s_channels: int,
        num_blocks: int,
        num_heads: int,
        nonlinearity: str = "gelu",
        mlp_ratio: int = 2,
        attn_ratio: int = 1,
        num_layers_mlp: int = 2,
        dropout_prob: float | None = None,
        checkpoint_blocks: bool = False,
        compile: bool = False,
    ):
        """
        Parameters
        ----------
        in_v_channels : int
            Number of input vector channels.
        out_v_channels : int
            Number of output vector channels.
        hidden_v_channels : int
            Number of hidden vector channels.
        in_s_channels : int
            Number of input scalar channels.
        out_s_channels : int
            Number of output scalar channels.
        hidden_s_channels : int
            Number of hidden scalar channels.
        num_blocks : int
            Number of Lorentz Transformer blocks.
        num_heads : int
            Number of attention heads.
        nonlinearity : str, optional
            Nonlinearity type for MLP layers, by default "gelu".
        mlp_ratio : int, optional
            Expansion ratio for MLP hidden layers, by default 2.
        attn_ratio : int, optional
            Expansion ratio for attention hidden layers, by default 1.
        num_layers_mlp : int, optional
            Number of layers in MLP, by default 2.
        dropout_prob : float | None, optional
            Dropout probability, by default None.
        checkpoint_blocks : bool, optional
            Whether to use gradient checkpointing for blocks, by default False.
        compile : bool, optional
            Whether to compile the model with torch.compile, by default False.
        """
        super().__init__()

        self.linear_in = Linear(
            in_v_channels=in_v_channels,
            in_s_channels=in_s_channels,
            out_v_channels=hidden_v_channels,
            out_s_channels=hidden_s_channels,
        )

        self.blocks = nn.ModuleList(
            [
                LGATrSlimBlock(
                    v_channels=hidden_v_channels,
                    s_channels=hidden_s_channels,
                    num_heads=num_heads,
                    nonlinearity=nonlinearity,
                    mlp_ratio=mlp_ratio,
                    attn_ratio=attn_ratio,
                    num_layers_mlp=num_layers_mlp,
                    dropout_prob=dropout_prob,
                )
                for _ in range(num_blocks)
            ]
        )

        self.linear_out = Linear(
            in_v_channels=hidden_v_channels,
            in_s_channels=hidden_s_channels,
            out_v_channels=out_v_channels,
            out_s_channels=out_s_channels,
        )
        self._checkpoint_blocks = checkpoint_blocks

        self.compile = compile
        if compile:
            # ugly hack to make torch.compile convenient for users
            # the clean solution is model = torch.compile(model, **kwargs) outside of the constructor
            # note that we need fullgraph=False because of the torch.compiler.disable for attention
            self.__class__ = torch.compile(self.__class__, dynamic=True, mode="default")

    def forward(self, vectors, scalars, **attn_kwargs):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.
        **attn_kwargs : dict
            Additional keyword arguments for the attention function.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        h_v, h_s = self.linear_in(vectors, scalars)

        for block in self.blocks:
            if self._checkpoint_blocks:
                h_v, h_s = checkpoint(block, h_v, h_s, use_reentrant=False, **attn_kwargs)
            else:
                h_v, h_s = block(h_v, h_s, **attn_kwargs)

        outputs_v, outputs_s = self.linear_out(h_v, h_s)
        return outputs_v, outputs_s



"""L-GATr self-attention."""

import torch
from einops import rearrange
from torch import nn

from ..dropout import GradeDropout
from ..linear import EquiLinear
from .attention import GeometricAttention
from .config import SelfAttentionConfig
from .qkv import MultiQueryQKVModule, QKVModule


class SelfAttention(nn.Module):
    """L-GATr self-attention.

    Constructs queries, keys, and values, computes attention, and projects linearly to outputs.

    Parameters
    ----------
    config : SelfAttentionConfig
        Attention configuration.
    """

    def __init__(self, config: SelfAttentionConfig) -> None:
        super().__init__()

        # Store settings
        self.config = config

        # QKV computation
        self.qkv_module = MultiQueryQKVModule(config) if config.multi_query else QKVModule(config)

        # Output projection
        self.out_linear = EquiLinear(
            in_mv_channels=config.hidden_mv_channels * config.num_heads,
            out_mv_channels=config.out_mv_channels,
            in_s_channels=(
                None
                if config.in_s_channels is None
                else config.hidden_s_channels * config.num_heads
            ),
            out_s_channels=config.out_s_channels,
            initialization=config.output_init,
        )

        # Attention
        self.attention = GeometricAttention(config)

        # Dropout
        self.dropout: nn.Module | None
        if config.dropout_prob is not None:
            self.dropout = GradeDropout(config.dropout_prob)
        else:
            self.dropout = None

        # HeadScaleMHA
        self.use_head_scale = config.head_scale
        if self.use_head_scale:
            self.head_scale = nn.Parameter(torch.ones(config.num_heads))

    def forward(
        self,
        multivectors: torch.Tensor,
        additional_qk_features_mv: torch.Tensor | None = None,
        scalars: torch.Tensor | None = None,
        additional_qk_features_s: torch.Tensor | None = None,
        **attn_kwargs,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Computes self-attention.

        The result is the following:

        .. code-block::

            # For each head
            queries = linear_channels(inputs)
            keys = linear_channels(inputs)
            values = linear_channels(inputs)
            hidden = attention_items(queries, keys, values, biases=biases)
            head_output = linear_channels(hidden)

            # Combine results
            output = concatenate_heads head_output

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., items, mv_channels, 16).
        additional_qk_features_mv : None or torch.Tensor
            Additional multivector Q/K features with shape (..., items, add_qk_mv_channels, 16)
        scalars : None or torch.Tensor
            Optional input scalars with shape (..., items, num_items, s_channels)
        additional_qk_features_s : None or torch.Tensor
            Additional scalar Q/K features with shape (..., items, add_qk_mv_channels, 16)
        scalars : None or torch.Tensor
            Optional input scalars with shape (..., items, s_channels).
        **attn_kwargs
            Optional keyword arguments passed to attention.

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., items, mv_channels, 16).
        output_scalars : torch.Tensor
            Output scalars with shape (..., items, s_channels).
        """
        # Compute Q, K, V
        q_mv, k_mv, v_mv, q_s, k_s, v_s = self.qkv_module(
            multivectors, scalars, additional_qk_features_mv, additional_qk_features_s
        )

        # Attention layer
        h_mv, h_s = self.attention(
            q_mv,
            k_mv,
            v_mv,
            q_s,
            k_s,
            v_s,
            **attn_kwargs,
        )
        if self.use_head_scale:
            h_mv = h_mv * self.head_scale.view(
                *[1] * len(h_mv.shape[:-5]), len(self.head_scale), 1, 1, 1
            )
            h_s = h_s * self.head_scale.view(*[1] * len(h_s.shape[:-4]), len(self.head_scale), 1, 1)

        h_mv = rearrange(
            h_mv,
            "... n_heads n_items hidden_channels x -> ... n_items (n_heads hidden_channels) x",
        )
        h_s = rearrange(
            h_s,
            "... n_heads n_items hidden_channels -> ... n_items (n_heads hidden_channels)",
        )

        # Transform linearly one more time
        outputs_mv, outputs_s = self.out_linear(h_mv, scalars=h_s)

        # Dropout
        if self.dropout is not None:
            outputs_mv, outputs_s = self.dropout(outputs_mv, outputs_s)

        return outputs_mv, outputs_s



"""Original flash-attention backend."""

import torch

try:
    from flash_attn import flash_attn_varlen_func
except ModuleNotFoundError as err:
    raise ImportError(
        "flash-attn is not installed. Run 'pip install lgatr[flash-attention]'."
    ) from err


@torch.compiler.disable()
def attention(query, key, value, dtype=None, **kwargs):
    """Pass to flash-attention's flash_attn_varlen_func.
    Note that flash-attention expects the shape (batch=1, items, head, channel).

    There is no fancy docs website, so one has to check the source code for the interface:
    https://github.com/Dao-AILab/flash-attention/blob/main/flash_attn/flash_attn_interface.py

    Parameters
    ----------
    query : torch.Tensor
        Queries with shape (batch, head, items_out, channel)
    key : torch.Tensor
        Keys with shape (batch, head, items_in, channel)
    value : torch.Tensor
        Values with shape (batch, head, items_in, channel)
    dtype : torch.dtype, optional
        If specified, cast input tensors to this dtype before passing to flash-attention.
        If None, use torch.get_autocast_gpu_dtype().
    **kwargs
        Additional keyword arguments passed to flash_attn_varlen_func.

    Returns
    -------
    out : torch.Tensor
        Result with shape (batch, head, items_out, channel)
    """
    assert len(query.shape) == 4, (
        "flash-attn constrains attention input shape to (batch, head, items, channel)."
    )

    if query.dtype not in [torch.float16, torch.bfloat16]:
        # flash-attention only supports fp16 and bf16
        if dtype is None:
            dtype = torch.get_autocast_gpu_dtype()
        in_dtype = query.dtype
        query, key, value = query.to(dtype), key.to(dtype), value.to(dtype)
    else:
        in_dtype = None

    def reshape(x):
        assert x.shape[0] == 1
        return x.squeeze(0).transpose(0, 1).contiguous()

    query, key, value = reshape(query), reshape(key), reshape(value)
    out = flash_attn_varlen_func(query, key, value, **kwargs)
    out = out.transpose(0, 1).unsqueeze(0).contiguous()

    if in_dtype is not None:
        out = out.to(in_dtype)
    return out



import pytest
import torch

from lgatr.layers import CrossAttention, CrossAttentionConfig
from tests.helpers import BATCH_DIMS, MILD_TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("items,items_condition", [(3, 8)])
@pytest.mark.parametrize(
    "in_kv_mv_channels,in_q_mv_channels,in_kv_s_channels,in_q_s_channels",
    [(2, 3, 4, 5)],
)
@pytest.mark.parametrize("multi_query,head_scale,", [(True, True), (False, False)])
@pytest.mark.parametrize("num_heads,increase_hidden_channels", [(3, 2)])
@pytest.mark.parametrize("dropout_prob", [None])
def test_crossattention_equivariance(
    batch_dims,
    items,
    items_condition,
    in_kv_mv_channels,
    in_q_mv_channels,
    in_kv_s_channels,
    in_q_s_channels,
    multi_query,
    head_scale,
    num_heads,
    increase_hidden_channels,
    dropout_prob,
):
    """Test cross attention equivariance."""

    config = CrossAttentionConfig(
        in_kv_mv_channels=in_kv_mv_channels,
        in_q_mv_channels=in_q_mv_channels,
        out_mv_channels=in_q_mv_channels,
        in_kv_s_channels=in_kv_s_channels,
        in_q_s_channels=in_q_s_channels,
        out_s_channels=in_q_s_channels,
        num_heads=num_heads,
        head_scale=head_scale,
        increase_hidden_channels=increase_hidden_channels,
        multi_query=multi_query,
        dropout_prob=dropout_prob,
    )
    layer = CrossAttention(config)

    scalars = torch.randn(*batch_dims, items, in_q_s_channels)
    scalars_condition = torch.randn(*batch_dims, items_condition, in_kv_s_channels)

    data_dims = [
        tuple(list(batch_dims) + [items_condition, in_kv_mv_channels]),
        tuple(list(batch_dims) + [items, in_q_mv_channels]),
    ]
    check_pin_equivariance(
        layer,
        2,
        batch_dims=data_dims,
        fn_kwargs=dict(scalars_kv=scalars_condition, scalars_q=scalars),
        **MILD_TOLERANCES,
    )



from .attention import sdp_attention
from .bilinear import geometric_product
from .config import gatr_config
from .dropout import grade_dropout
from .invariants import abs_squared_norm, inner_product
from .linear import equi_linear, grade_involute, grade_project, reverse
from .nonlinearities import gated_gelu, gated_relu, gated_sigmoid, gated_silu
from .normalization import equi_layer_norm



"""Self-attention layers."""

from torch import nn

from ...primitives.attention import sdp_attention
from .config import SelfAttentionConfig


class GeometricAttention(nn.Module):
    """Geometric attention layer.

    This is the main attention mechanism used in L-GATr.

    Given multivector and scalar queries, keys, and values, this layer computes:

    .. code-block::

        attn_weights[..., i, j] = softmax_j[
            ga_inner_product(q_mv[..., i, :, :], k_mv[..., j, :, :])
            + euclidean_inner_product(q_s[..., i, :], k_s[..., j, :])
        ]
        out_mv[..., i, c, :] = sum_j attn_weights[..., i, j] v_mv[..., j, c, :] / norm
        out_s[..., i, c] = sum_j attn_weights[..., i, j] v_s[..., j, c] / norm

    Parameters
    ----------
    config : SelfAttentionConfig
        Attention configuration.
    """

    def __init__(self, config: SelfAttentionConfig) -> None:
        super().__init__()

    def forward(self, q_mv, k_mv, v_mv, q_s, k_s, v_s, **attn_kwargs):
        """Forward pass through geometric attention.

        Given multivector and scalar queries, keys, and values, this forward pass computes:

        .. code-block::

            attn_weights[..., i, j] = softmax_j[
                ga_inner_product(q_mv[..., i, :, :], k_mv[..., j, :, :])
                + euclidean_inner_product(q_s[..., i, :], k_s[..., j, :])
            ]
            out_mv[..., i, c, :] = sum_j attn_weights[..., i, j] v_mv[..., j, c, :] / norm
            out_s[..., i, c] = sum_j attn_weights[..., i, j] v_s[..., j, c] / norm

        Parameters
        ----------
        q_mv : torch.Tensor
            Multivector queries with shape (..., items_out, mv_channels, 16).
        k_mv : torch.Tensor
            Multivector keys with shape (..., items_in, mv_channels, 16).
        v_mv : torch.Tensor
            Multivector values with shape (..., items_in, mv_channels, 16).
        q_s : torch.Tensor
            Scalar queries with shape (..., items_out, s_channels).
        k_s : torch.Tensor
            Scalar keys with shape (..., items_in, s_channels).
        v_s : torch.Tensor
            Scalar values with shape (..., items_in, s_channels).
        **attn_kwargs
            Optional keyword arguments passed to attention.
        """

        h_mv, h_s = sdp_attention(
            q_mv,
            k_mv,
            v_mv,
            q_s,
            k_s,
            v_s,
            **attn_kwargs,
        )

        return h_mv, h_s



"""Unit tests of bilinear primitives."""

import pytest
import torch

from lgatr.primitives.config import gatr_config
from lgatr.primitives.linear import equi_linear, grade_involute, grade_project, reverse
from tests.helpers import (
    BATCH_DIMS,
    TOLERANCES,
    check_consistence_with_grade_involution,
    check_consistence_with_reversal,
    check_pin_equivariance,
)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_reverse_correctness(batch_dims):
    """Tests the multivector reverse for equivariance."""
    check_consistence_with_reversal(reverse, batch_dims=batch_dims, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_grade_involution_correctness(batch_dims):
    """Tests the multivector reverse for equivariance."""
    check_consistence_with_grade_involution(grade_involute, batch_dims=batch_dims, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_identity_equivariance(batch_dims):
    """Tests an identity map for equivariance (testing the test)."""
    check_pin_equivariance(lambda x: x, 1, batch_dims=batch_dims, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_grade_project_equivariance(batch_dims):
    """Tests the grade_project() primitive for equivariance."""
    check_pin_equivariance(grade_project, 1, batch_dims=batch_dims, **TOLERANCES)


@pytest.mark.parametrize(
    "input_batch_dims,coeff_batch_dims",
    [
        ((7,), (5, 7)),
        ((3, 7), (5, 7)),
        ((2, 3, 7), (5, 7)),
    ],
)
def test_linear_equivariance(input_batch_dims, coeff_batch_dims):
    """Tests the equi_linear() primitive for equivariance."""
    fn_kwargs = dict(
        coeffs=torch.randn(*coeff_batch_dims, gatr_config.num_pin_linear_basis_elements),
    )
    check_pin_equivariance(
        equi_linear, 1, fn_kwargs=fn_kwargs, batch_dims=input_batch_dims, **TOLERANCES
    )



import pytest
import torch

from lgatr.nets.lgatr_slim import (
    MLP,
    Dropout,
    GatedLinearUnit,
    LGATrSlim,
    LGATrSlimBlock,
    Linear,
    RMSNorm,
    SelfAttention,
    squared_norm,
)

from ...helpers.constants import BATCH_DIMS, TOLERANCES
from ...helpers.equivariance_noga import check_equivariance, check_invariance

CHANNELS = [
    (5, 1, 4, 2),
    (1, 4, 0, 2),
    (9, 3, 4, 0),
    (2, 7, 0, 0),
    (0, 1, 2, 3),
    (3, 0, 2, 3),
    (0, 0, 2, 3),
]


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_squared_norm_invariance(batch_dims):
    check_invariance(squared_norm, batch_dims=batch_dims, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("dropout_prob", [0.0, 0.1, 0.5])
def test_Dropout_equivariance(batch_dims, dropout_prob):
    layer = Dropout(dropout_prob)
    layer.eval()

    # shape
    v = torch.randn(*batch_dims, 4)
    s = torch.randn(*batch_dims)
    out_v, out_s = layer(v, scalars=s)
    assert out_v.shape == v.shape
    assert out_s.shape == s.shape

    # equivariance
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_RMSNorm_equivariance(batch_dims):
    layer = RMSNorm()

    # shape
    v = torch.randn(*batch_dims, 4)
    s = torch.randn(*batch_dims)
    out_v, out_s = layer(v, scalars=s)
    assert out_v.shape == v.shape
    assert out_s.shape == s.shape

    # equivariance
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("nonlinearity", ["relu", "sigmoid", "tanh", "gelu", "silu"])
@pytest.mark.parametrize("in_v_channels,out_v_channels,in_s_channels,out_s_channels", CHANNELS)
def test_GatedLinearUnit_equivariance(
    batch_dims, nonlinearity, in_v_channels, out_v_channels, in_s_channels, out_s_channels
):
    layer = GatedLinearUnit(
        in_v_channels=in_v_channels,
        out_v_channels=out_v_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
        nonlinearity=nonlinearity,
    )
    s = torch.randn(*batch_dims, in_s_channels)
    v = torch.randn(*batch_dims, in_v_channels, 4)
    out_v, out_s = layer(v, s)
    assert out_v.shape == v.shape[:-2] + (out_v_channels, 4)
    assert out_s.shape == s.shape[:-1] + (out_s_channels,)

    # equivariance
    batch_dims = batch_dims + [in_v_channels]
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("in_v_channels,out_v_channels,in_s_channels,out_s_channels", CHANNELS)
@pytest.mark.parametrize("initialization", ["default", "small"])
def test_Linear_equivariance(
    batch_dims,
    in_v_channels,
    out_v_channels,
    in_s_channels,
    out_s_channels,
    initialization,
):
    layer = Linear(
        in_v_channels=in_v_channels,
        out_v_channels=out_v_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
        initialization=initialization,
    )
    s = torch.randn(*batch_dims, in_s_channels)
    v = torch.randn(*batch_dims, in_v_channels, 4)
    out_v, out_s = layer(v, s)
    assert out_v.shape == v.shape[:-2] + (out_v_channels, 4)
    assert out_s.shape == s.shape[:-1] + (out_s_channels,)

    # equivariance
    batch_dims = batch_dims + [in_v_channels]
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", [(100,)])
@pytest.mark.parametrize("in_v_channels,out_v_channels,in_s_channels,out_s_channels", CHANNELS[:4])
def test_Linear_initialization(
    batch_dims,
    in_v_channels,
    out_v_channels,
    in_s_channels,
    out_s_channels,
    var_tolerance=10.0,
):
    # Test that inputs with variance 1 are roughly mapped to outputs with variance 1
    layer = Linear(
        in_v_channels=in_v_channels,
        out_v_channels=out_v_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
    )

    inputs_v = torch.randn(*batch_dims, in_v_channels, 4)
    inputs_s = torch.randn(*batch_dims, in_s_channels)
    outputs_v, outputs_s = layer(inputs_v, inputs_s)

    v_mean = outputs_v.cpu().detach().to(torch.float64).mean(dim=(0, 1))
    v_var = outputs_v.cpu().detach().to(torch.float64).var(dim=(0, 1))
    target_mean = torch.zeros_like(v_mean)
    target_var = torch.ones_like(v_var) / 3.0
    assert torch.all(v_mean > target_mean - 0.3)
    assert torch.all(v_mean < target_mean + 0.3)
    assert torch.all(v_var > target_var / var_tolerance)
    assert torch.all(v_var < target_var * var_tolerance)

    if out_s_channels > 0 and in_s_channels > 0:
        s_mean = outputs_s.cpu().detach().to(torch.float64).mean().item()
        s_var = outputs_s.cpu().detach().to(torch.float64).var().item()

        assert -1.0 < s_mean < 1.0
        assert 1.0 / var_tolerance < s_var < 1.0 * var_tolerance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("v_channels,s_channels", [(24, 14)])
@pytest.mark.parametrize("num_heads,attn_ratio", [(2, 1), (1, 2)])
def test_SelfAttention_equivariance(
    batch_dims,
    v_channels,
    s_channels,
    num_heads,
    attn_ratio,
):
    layer = SelfAttention(
        v_channels=v_channels,
        s_channels=s_channels,
        num_heads=num_heads,
        attn_ratio=attn_ratio,
    )
    s = torch.randn(*batch_dims, s_channels)

    v = torch.randn(*batch_dims, v_channels, 4)
    out_v, out_s = layer(v, s)
    assert out_v.shape == v.shape
    assert out_s.shape == s.shape

    batch_dims = batch_dims + [v_channels]
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("v_channels,s_channels", [(32, 4), (16, 8)])
@pytest.mark.parametrize("mlp_ratio,num_layers", [(1, 2), (2, 2), (1, 3)])
def test_MLP_equivariance(batch_dims, v_channels, s_channels, mlp_ratio, num_layers):
    layer = MLP(
        v_channels=v_channels,
        s_channels=s_channels,
        mlp_ratio=mlp_ratio,
        num_layers=num_layers,
    )
    s = torch.randn(*batch_dims, s_channels)
    batch_dims = batch_dims + [v_channels]

    # equivariance
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("v_channels,s_channels,num_heads", [(32, 4, 1), (16, 8, 4)])
@pytest.mark.parametrize("dropout_prob", [None, 0.0, 0.5])
def test_LGATrSlimBlock_equivariance(
    batch_dims,
    v_channels,
    s_channels,
    num_heads,
    dropout_prob,
):
    layer = LGATrSlimBlock(
        v_channels=v_channels,
        s_channels=s_channels,
        num_heads=num_heads,
        dropout_prob=dropout_prob,
    )
    layer.eval()
    s = torch.randn(*batch_dims, s_channels)
    batch_dims = batch_dims + [v_channels]

    # equivariance
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize(
    "in_v_channels,in_s_channels,out_v_channels,out_s_channels",
    [
        (4, 3, 9, 2),
        (2, 9, 0, 3),
        (3, 5, 7, 0),
        (8, 3, 0, 0),
    ],
)
@pytest.mark.parametrize("hidden_v_channels,hidden_s_channels,num_heads", [(32, 4, 1), (16, 8, 4)])
@pytest.mark.parametrize("dropout_prob", [None, 0.0, 0.5])
@pytest.mark.parametrize("num_blocks", [1, 2])
@pytest.mark.parametrize("checkpoint_blocks", [False, True])
def test_LGATrSlim_equivariance(
    batch_dims,
    in_v_channels,
    in_s_channels,
    out_v_channels,
    out_s_channels,
    hidden_v_channels,
    hidden_s_channels,
    num_heads,
    num_blocks,
    dropout_prob,
    checkpoint_blocks,
):
    layer = LGATrSlim(
        in_v_channels=in_v_channels,
        out_v_channels=out_v_channels,
        hidden_v_channels=hidden_v_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
        hidden_s_channels=hidden_s_channels,
        num_blocks=num_blocks,
        num_heads=num_heads,
        dropout_prob=dropout_prob,
        checkpoint_blocks=checkpoint_blocks,
    )
    layer.eval()
    s = torch.randn(*batch_dims, in_s_channels)
    v = torch.randn(*batch_dims, in_v_channels, 4)
    out_v, out_s = layer(v, s)
    assert out_v.shape == v.shape[:-2] + (out_v_channels, 4)
    assert out_s.shape == s.shape[:-1] + (out_s_channels,)

    # equivariance
    batch_dims = batch_dims + [in_v_channels]
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize(
    "in_v_channels,in_s_channels,out_v_channels,out_s_channels", [(4, 3, 9, 2)]
)
@pytest.mark.parametrize(
    "hidden_v_channels,hidden_s_channels,num_heads,num_blocks", [(16, 8, 4, 1)]
)
def test_LGATrSlim_equivariance_compiled(
    batch_dims,
    in_v_channels,
    in_s_channels,
    out_v_channels,
    out_s_channels,
    hidden_v_channels,
    hidden_s_channels,
    num_heads,
    num_blocks,
    compile=True,
):
    layer = LGATrSlim(
        in_v_channels=in_v_channels,
        out_v_channels=out_v_channels,
        hidden_v_channels=hidden_v_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
        hidden_s_channels=hidden_s_channels,
        num_blocks=num_blocks,
        num_heads=num_heads,
        compile=compile,
    )
    layer.eval()
    s = torch.randn(*batch_dims, in_s_channels)
    v = torch.randn(*batch_dims, in_v_channels, 4)
    out_v, out_s = layer(v, s)
    assert out_v.shape == v.shape[:-2] + (out_v_channels, 4)
    assert out_s.shape == s.shape[:-1] + (out_s_channels,)

    # equivariance
    batch_dims = batch_dims + [in_v_channels]
    check_equivariance(layer, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)



"""Utility functions to test callables for equivariance with respect to Pin(1,3)."""

import torch

from .clifford import SlowRandomPinTransform


def get_first_output(outputs):
    """Extracts the first output of a tuple of multiple outputs.

    If only one output is present, returns that.

    This is convenient for equivariance checks: primitives usually just return one multivector
    tensor, while layers return two tensors. This function can be wrapped around a callable that's
    either a primitive or a layer and will always return the (first) multivector output.
    """
    if isinstance(outputs, tuple):
        return outputs[0]

    return outputs


def check_pin_equivariance(
    function,
    num_multivector_args=1,
    fn_kwargs=None,
    batch_dims=(1,),
    spin=True,
    rng=None,
    num_checks=3,
    **kwargs,
):
    """Checks whether a callable is equivariant with respect to the Pin(1,3) or Spin(1,3) group.

    The callable can have an arbitrary number of multivector inputs.

    Parameters
    ----------
    function: Callable
        Function to be tested for equivariance. The first `num_multivector_args` positional
        arguments need to accept torch.Tensor inputs describing multivectors, and will be
        transformed as part of the equivariance test.
    num_multivector_args: int
        Number of multivector that `function` accepts.
    fn_kwargs : dict with str keys
        Keyword arguments to call `function` with.
    batch_dims : Tuple[int] or List[Tuple[int]]
        Batch shapes for the multivector inputs to `function`.
        Expects List[Tuple[int]] if num_multivector_args==1
    spin : bool
        If True, this function tests Spin equivariance; if False, it tests Pin equivariance.
    rng : numpy.random.Generator or None
        Numpy rng to draw the inputs and transformations from.
    num_checks : int
        Number of function calls (with random inputs) to determine whether the function passes the
        equivariance test.
    kwargs
        Optional keyword arguments for the equality check. Will be passed on to np.allclose.
        This can for instance be used to specify the absolute and relative tolerance
        (by passing `atol` and `rtol` keyword arguments).
    """
    # Default arguments
    if fn_kwargs is None:
        fn_kwargs = {}

    # Propagate numpy random state to torch
    if rng is not None:
        torch.manual_seed(rng.integers(100000))

    if num_multivector_args == 1:
        batch_dims = [batch_dims]
    assert num_multivector_args == len(batch_dims)

    # Loop over multiple checks
    for _ in range(num_checks):
        # Generate function inputs and Pin(3,0,1) transformations
        # Generate function inputs
        inputs = [torch.randn(*batch_dim, 16) for batch_dim in batch_dims]

        transform = SlowRandomPinTransform(rng=rng, spin=spin)

        # First function, then transformation
        outputs = get_first_output(function(*inputs, **fn_kwargs))
        transformed_outputs = transform(outputs)

        # First transformation, then function
        transformed_inputs = [transform(inputss) for inputss in inputs]
        outputs_of_transformed = get_first_output(function(*transformed_inputs, **fn_kwargs))

        # Check equality
        torch.testing.assert_close(transformed_outputs, outputs_of_transformed, **kwargs)


def check_pin_invariance(
    function,
    num_multivector_args=1,
    fn_kwargs=None,
    batch_dims=(1,),
    spin=True,
    rng=None,
    num_checks=3,
    **kwargs,
):
    """Checks whether a callable is invariant with respect to the Pin(1,3) or Spin(1,3) group.

    Parameters
    ----------
    function: Callable
        Function to be tested for equivariance. The first `num_multivector_args` positional
        arguments need to accept torch.Tensor inputs describing multivectors, and will be
        transformed as part of the invariance test.
    num_multivector_args: int
        Number of multivector that `function` accepts.
    fn_kwargs : dict with str keys
        Keyword arguments to call `function` with.
    batch_dims : Tuple[int] or List[Tuple[int]]
        Batch shapes for the multivector inputs to `function`.
        Expects List[Tuple[int]] if num_multivector_args==1
    spin : bool
        If True, this function tests Spin equivariance; if False, it tests Pin equivariance.
        Since Spin is a subgroup of Pin, it is usually enough to confirm Pin equivariance.
    rng : numpy.random.Generator or None
        Numpy rng to draw the inputs and transformations from.
    num_checks : int
        Number of function calls (with random inputs) to determine whether the function passes the
        equivariance test.
    kwargs
        Optional keyword arguments for the equality check. Will be passed on to np.allclose.
        This can for instance be used to specify the absolute and relative tolerance
        (by passing `atol` and `rtol` keyword arguments).
    """
    # Default arguments
    if fn_kwargs is None:
        fn_kwargs = {}

    # Propagate numpy random state to torch
    if rng is not None:
        torch.manual_seed(rng.integers(100000))

    if num_multivector_args == 1:
        batch_dims = [batch_dims]
    assert num_multivector_args == len(batch_dims)

    # Loop over multiple checks
    for _ in range(num_checks):
        # Generate function inputs
        inputs = [torch.randn(*batch_dim, 16) for batch_dim in batch_dims]

        # Transform inputs with Pin(1,3)
        transform = SlowRandomPinTransform(rng=rng, spin=spin)
        transformed_inputs = [transform(inputss) for inputss in inputs]

        # Evaluate function on original and transformed inputs
        outputs = get_first_output(function(*inputs, **fn_kwargs))
        outputs_of_transformed = get_first_output(function(*transformed_inputs, **fn_kwargs))

        # Check equality
        torch.testing.assert_close(outputs, outputs_of_transformed, **kwargs)



"""Utility functions to test callables for consistency with clifford algebras."""

from math import prod

import clifford
import torch

from .clifford import mv_list_to_tensor

LAYOUT, _ = clifford.Cl(1, 3)


def _sample_list_of_mv(batch_dims, rng):
    """Utility function that samples a list of multivectors."""
    total_batchsize = 1 if not batch_dims else prod(batch_dims)
    xs = clifford.randomMV(layout=LAYOUT, n=total_batchsize, rng=rng)
    if total_batchsize == 1:  # Dealing with inconsistency of clifford.randomMV
        xs = [xs]
    return xs


def check_consistence_with_geometric_product(function, batch_dims=(1,), rng=None, **kwargs):
    """Checks whether a callable computes the geometric product."""

    # Generate random inputs
    xs = _sample_list_of_mv(batch_dims, rng)
    ys = _sample_list_of_mv(batch_dims, rng)

    # Compute ground-truth geometric product
    xy_true = [(x * y) for x, y in zip(xs, ys, strict=False)]
    xy_true = mv_list_to_tensor(xy_true, batch_dims)

    # Compute geometric product with function
    x_tensor = mv_list_to_tensor(xs, batch_dims)
    y_tensor = mv_list_to_tensor(ys, batch_dims)
    xy = function(x_tensor, y_tensor)

    # Compare
    torch.testing.assert_close(xy, xy_true, **kwargs)


def check_consistence_with_reversal(function, batch_dims=(1,), rng=None, **kwargs):
    """Checks whether a callable computes the reversal of a multivector."""

    # Generate random inputs
    xs = _sample_list_of_mv(batch_dims, rng)

    # Compute ground-truth reversal
    reversed_x_true = [~x for x in xs]
    reversed_x_true = mv_list_to_tensor(reversed_x_true, batch_dims)

    # Compute reversal with function
    x_tensor = mv_list_to_tensor(xs, batch_dims)
    reversed_x = function(x_tensor)

    # Compare
    torch.testing.assert_close(reversed_x, reversed_x_true, **kwargs)


def check_consistence_with_grade_involution(function, batch_dims=(1,), rng=None, **kwargs):
    """Checks whether a callable computes the grade involution of a multivectors."""

    # Generate random inputs
    xs = _sample_list_of_mv(batch_dims, rng)

    # Compute ground-truth reversal
    reversed_x_true = [x.gradeInvol() for x in xs]
    reversed_x_true = mv_list_to_tensor(reversed_x_true, batch_dims)

    # Compute reversal with function
    x_tensor = mv_list_to_tensor(xs, batch_dims)
    reversed_x = function(x_tensor)

    # Compare
    torch.testing.assert_close(reversed_x, reversed_x_true, **kwargs)



"""L-GATr decoder block."""

from dataclasses import replace

import torch
from torch import nn

from .attention import (
    CrossAttention,
    CrossAttentionConfig,
    SelfAttention,
    SelfAttentionConfig,
)
from .layer_norm import EquiLayerNorm
from .mlp.config import MLPConfig
from .mlp.mlp import GeoMLP


class ConditionalLGATrBlock(nn.Module):
    """L-GATr decoder block.

    Inputs are first processed by a block consisting of LayerNorm, multi-head geometric
    self-attention, and a residual connection. Then the conditions are included with
    cross-attention using the same overhead as in the self-attention part.
    Then the data is processed by a block consisting of another LayerNorm,
    an item-wise two-layer geometric MLP with GeLU activations, and another
    residual connection.

    Parameters
    ----------
    mv_channels : int
        Number of input and output multivector channels
    s_channels: int
        Number of input and output scalar channels
    condition_mv_channels: int
        Number of condition multivector channels
    condition_s_channels: int
        Number of condition scalar channels
    attention: SelfAttentionConfig
        Attention configuration
    crossattention: CrossAttentionConfig
        Cross-attention configuration
    mlp: MLPConfig
        MLP configuration
    dropout_prob : float or None
        Dropout probability
    """

    def __init__(
        self,
        mv_channels: int,
        s_channels: int,
        condition_mv_channels: int,
        condition_s_channels: int,
        attention: SelfAttentionConfig,
        crossattention: CrossAttentionConfig,
        mlp: MLPConfig,
        dropout_prob: float | None = None,
    ) -> None:
        super().__init__()

        # Normalization layer (stateless, so we can use the same layer for both normalization instances)
        self.norm = EquiLayerNorm()

        # Self-attention layer
        attention = replace(
            attention,
            in_mv_channels=mv_channels,
            out_mv_channels=mv_channels,
            in_s_channels=s_channels,
            out_s_channels=s_channels,
            output_init="small",
            dropout_prob=dropout_prob,
        )
        self.attention = SelfAttention(attention)

        # Cross-attention layer
        crossattention = replace(
            crossattention,
            in_q_mv_channels=mv_channels,
            in_q_s_channels=s_channels,
            in_kv_mv_channels=condition_mv_channels,
            in_kv_s_channels=condition_s_channels,
            out_mv_channels=mv_channels,
            out_s_channels=s_channels,
            output_init="small",
            dropout_prob=dropout_prob,
        )
        self.crossattention = CrossAttention(crossattention)

        # MLP block
        mlp = replace(
            mlp,
            mv_channels=mv_channels,
            s_channels=s_channels,
            dropout_prob=dropout_prob,
        )
        self.mlp = GeoMLP(mlp)

    def forward(
        self,
        multivectors: torch.Tensor,
        multivectors_condition: torch.Tensor,
        scalars: torch.Tensor = None,
        scalars_condition: torch.Tensor = None,
        attn_kwargs=None,
        crossattn_kwargs=None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of the transformer decoder block.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors  with shape (..., items, mv_channels, 16).
        scalars : torch.Tensor
            Input scalars with shape (..., items, s_channels).
        multivectors_condition : torch.Tensor
            Input condition multivectors with shape (..., items, mv_channels, 16).
        scalars_condition : torch.Tensor
            Input condition scalars with shape (..., items, s_channels).
        attn_kwargs: None or torch.Tensor or AttentionBias
            Optional attention mask.
        crossattn_kwargs: None or torch.Tensor or AttentionBias
            Optional attention mask for the condition.

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., items, mv_channels, 16).
        output_scalars : torch.Tensor
            Output scalars with shape (..., items, s_channels).
        """
        attn_kwargs = attn_kwargs if attn_kwargs is not None else {}
        crossattn_kwargs = crossattn_kwargs if crossattn_kwargs is not None else {}

        # Self-attention block: pre layer norm
        h_mv, h_s = self.norm(multivectors, scalars=scalars)

        # Self-attention block: self attention
        h_mv, h_s = self.attention(
            h_mv,
            scalars=h_s,
            **attn_kwargs,
        )

        # Self-attention block: skip connection
        multivectors = multivectors + h_mv
        scalars = scalars + h_s

        # Cross-attention block: pre layer norm
        h_mv, h_s = self.norm(multivectors, scalars=scalars)
        c_mv, c_s = self.norm(multivectors_condition, scalars=scalars_condition)

        # Cross-attention block: cross attention
        h_mv, h_s = self.crossattention(
            multivectors_q=h_mv,
            multivectors_kv=c_mv,
            scalars_q=h_s,
            scalars_kv=c_s,
            **crossattn_kwargs,
        )

        # Cross-attention block: skip connection
        outputs_mv = multivectors + h_mv
        outputs_s = scalars + h_s

        # MLP block: pre layer norm
        h_mv, h_s = self.norm(outputs_mv, scalars=outputs_s)

        # MLP block: MLP
        h_mv, h_s = self.mlp(h_mv, scalars=h_s)

        # MLP block: skip connection
        outputs_mv = outputs_mv + h_mv
        outputs_s = outputs_s + h_s

        return outputs_mv, outputs_s



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class LGATrConfig:
    """Configuration for global settings like the symmetry group.

    Parameters
    ----------
    use_fully_connected_subgroup : bool
        If True, model is only equivariant with respect to
        the fully connected subgroup of the Lorentz group,
        the proper orthochronous Lorentz group :math:`SO^+(1,3)`,
        which does not include parity and time reversal.
        This setting affects how the EquiLinear maps work:
        For :math:`SO^+(1,3)`, they include transitions scalars/pseudoscalars
        vectors/axialvectors and among bivectors, effectively
        treating the pseudoscalar/axialvector representations
        like another scalar/vector.
        Defaults to False, because parity-odd representations
        are usually not important in high-energy physics simulations.
    use_bivector : bool
        If False, the bivector components are set to zero after they
        are created in the GeometricBilinear layer.
        This is a toy switch to explore the effect of higher-order
        representations.
    use_geometric_product : bool
        If False, the GeometricBilinear layer is replaced
        by a EquiLinear + ScalarGatedNonlinearity layer.
        This is a toy switch to explore the effect of the geometric product.
    """

    use_fully_connected_subgroup: bool = True

    use_bivector: bool = True
    use_geometric_product: bool = True

    @property
    def num_pin_linear_basis_elements(self):
        return 10 if self.use_fully_connected_subgroup else 5


gatr_config = LGATrConfig()



"""Equivariant transformer for multivector data."""

from dataclasses import replace

import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from ..layers.attention.config import SelfAttentionConfig
from ..layers.lgatr_block import LGATrBlock
from ..layers.linear import EquiLinear
from ..layers.mlp.config import MLPConfig


class LGATr(nn.Module):
    """L-GATr network.

    It combines num_blocks L-GATr transformer blocks, each consisting of geometric self-attention
    layers, a geometric MLP, residual connections, and normalization layers. In addition, there
    are initial and final equivariant linear layers.

    Assumes input has shape (..., items, in_mv_channels, 16), output has shape
    (..., items, out_mv_channels, 16), will create hidden representations with shape
    (..., items, hidden_mv_channels, 16). Similar for extra scalar channels.

    Parameters
    ----------
    num_blocks : int
        Number of transformer blocks.
    in_mv_channels : int
        Number of input multivector channels.
    out_mv_channels : int
        Number of output multivector channels.
    hidden_mv_channels : int
        Number of hidden multivector channels.
    in_s_channels : None or int
        If not None, sets the number of scalar input channels.
    out_s_channels : None or int
        If not None, sets the number of scalar output channels.
    hidden_s_channels : None or int
        If not None, sets the number of scalar hidden channels.
    attention: Dict
        Data for SelfAttentionConfig
    mlp: Dict
        Data for MLPConfig
    reinsert_mv_channels : None or Tuple[int]
        If not None, specifies multivector channels that will be reinserted in every attention layer.
    reinsert_s_channels : None or Tuple[int]
        If not None, specifies scalar channels that will be reinserted in every attention layer.
    dropout_prob : float or None
        Dropout probability
    checkpoint_blocks : bool
        Whether to use checkpointing for the blocks. If True, will save memory at the cost of speed.
    """

    def __init__(
        self,
        num_blocks: int,
        in_mv_channels: int,
        out_mv_channels: int,
        hidden_mv_channels: int,
        in_s_channels: int | None,
        out_s_channels: int | None,
        hidden_s_channels: int | None,
        attention: SelfAttentionConfig,
        mlp: MLPConfig,
        reinsert_mv_channels: tuple[int] | None = None,
        reinsert_s_channels: tuple[int] | None = None,
        dropout_prob: float | None = None,
        checkpoint_blocks: bool = False,
    ) -> None:
        super().__init__()
        self.linear_in = EquiLinear(
            in_mv_channels,
            hidden_mv_channels,
            in_s_channels=in_s_channels,
            out_s_channels=hidden_s_channels,
        )
        attention = replace(
            SelfAttentionConfig.cast(attention),
            additional_qk_mv_channels=(
                0 if reinsert_mv_channels is None else len(reinsert_mv_channels)
            ),
            additional_qk_s_channels=0 if reinsert_s_channels is None else len(reinsert_s_channels),
        )
        mlp = MLPConfig.cast(mlp)
        self.blocks = nn.ModuleList(
            [
                LGATrBlock(
                    mv_channels=hidden_mv_channels,
                    s_channels=hidden_s_channels,
                    attention=attention,
                    mlp=mlp,
                    dropout_prob=dropout_prob,
                )
                for _ in range(num_blocks)
            ]
        )
        self.linear_out = EquiLinear(
            hidden_mv_channels,
            out_mv_channels,
            in_s_channels=hidden_s_channels,
            out_s_channels=out_s_channels,
        )
        self._reinsert_s_channels = reinsert_s_channels
        self._reinsert_mv_channels = reinsert_mv_channels
        self._checkpoint_blocks = checkpoint_blocks

    def forward(
        self,
        multivectors: torch.Tensor,
        scalars: torch.Tensor | None = None,
        **attn_kwargs,
    ) -> tuple[torch.Tensor, torch.Tensor | None]:
        """Forward pass of the network.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., items, in_mv_channels, 16).
        scalars : None or torch.Tensor
            Optional input scalars with shape (..., items, in_s_channels).
        **attn_kwargs
            Optional keyword arguments passed to attention.

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., items, out_mv_channels, 16).
        outputs_s : None or torch.Tensor
            Output scalars with shape (..., items, out_s_channels). None if out_s_channels=None.
        """

        # Channels that will be re-inserted in any query / key computation
        (
            additional_qk_features_mv,
            additional_qk_features_s,
        ) = self._construct_reinserted_channels(multivectors, scalars)

        # Pass through the blocks
        h_mv, h_s = self.linear_in(multivectors, scalars=scalars)
        for block in self.blocks:
            if self._checkpoint_blocks:
                h_mv, h_s = checkpoint(
                    block,
                    h_mv,
                    use_reentrant=False,
                    scalars=h_s,
                    additional_qk_features_mv=additional_qk_features_mv,
                    additional_qk_features_s=additional_qk_features_s,
                    **attn_kwargs,
                )
            else:
                h_mv, h_s = block(
                    h_mv,
                    scalars=h_s,
                    additional_qk_features_mv=additional_qk_features_mv,
                    additional_qk_features_s=additional_qk_features_s,
                    **attn_kwargs,
                )

        outputs_mv, outputs_s = self.linear_out(h_mv, scalars=h_s)

        return outputs_mv, outputs_s

    def _construct_reinserted_channels(self, multivectors, scalars):
        """Constructs input features that will be reinserted in every attention layer."""

        if self._reinsert_mv_channels is None:
            additional_qk_features_mv = None
        else:
            additional_qk_features_mv = multivectors[..., self._reinsert_mv_channels, :]

        if self._reinsert_s_channels is None:
            additional_qk_features_s = None
        else:
            assert scalars is not None
            additional_qk_features_s = scalars[..., self._reinsert_s_channels]

        return additional_qk_features_mv, additional_qk_features_s



import pytest
import torch

from lgatr.interface import embed_axialvector, extract_axialvector
from tests.helpers import BATCH_DIMS, TOLERANCES


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_axialvector_embedding_consistency(batch_dims):
    """Tests whether Lorentz vector embeddings into multivectors are cycle consistent."""
    axialvectors = torch.randn(*batch_dims, 4)
    multivectors = embed_axialvector(axialvectors)
    axialvectors_reencoded = extract_axialvector(multivectors)
    torch.testing.assert_close(axialvectors, axialvectors_reencoded, **TOLERANCES)



"""Geometric product."""

from functools import lru_cache
from pathlib import Path

import torch

from ..utils.einsum import cached_einsum
from .linear import DEFAULT_DEVICE, DEFAULT_DTYPE


@lru_cache
def _load_geometric_product_tensor(device=DEFAULT_DEVICE, dtype=DEFAULT_DTYPE) -> torch.Tensor:
    """Loads geometric product tensor for geometric product between multivectors.

    This function is cached.

    Parameters
    ----------
    device : torch.Device or str
        Device
    dtype : torch.Dtype
        Data type

    Returns
    -------
    basis : torch.Tensor
        Geometric product tensor with shape (16, 16, 16)
    """

    # To avoid duplicate loading, base everything on float32 CPU version
    if device not in [DEFAULT_DEVICE, "cpu"] and dtype != DEFAULT_DTYPE:
        gmt = _load_geometric_product_tensor()
    else:
        filename = Path(__file__).parent.resolve() / "geometric_product.pt"
        gmt = torch.load(filename).to(DEFAULT_DTYPE).to_dense()

    return gmt.to(device=device, dtype=dtype)


def geometric_product(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Computes the geometric product ``f(x,y) = x*y``.

    Parameters
    ----------
    x : torch.Tensor
        First input multivector with shape (..., 16).
        Batch dimensions must be broadcastable between x and y.
    y : torch.Tensor
        Second input multivector with shape (..., 16).
        Batch dimensions must be broadcastable between x and y.

    Returns
    -------
    outputs : torch.Tensor
        Result with shape (..., 16).
        Batch dimensions are result of broadcasting between x, y, and coeffs.
    """

    # Select kernel on correct device
    gp = _load_geometric_product_tensor(device=x.device, dtype=x.dtype)

    # Compute geometric product
    outputs = cached_einsum("i j k, ... j, ... k -> ... i", gp, x, y)

    return outputs



"""Unit tests of bilinear primitives."""

import pytest

from lgatr.primitives.bilinear import geometric_product
from tests.helpers import (
    BATCH_DIMS,
    TOLERANCES,
    check_consistence_with_geometric_product,
    check_pin_equivariance,
)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_geometric_product_correctness(batch_dims):
    """Tests the geometric_product() primitive for correctness (that is, consistency with the
    clifford library).
    """
    check_consistence_with_geometric_product(geometric_product, batch_dims, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_geometric_product_equivariance(batch_dims):
    """Tests the geometric_product() primitive for equivariance."""
    check_pin_equivariance(geometric_product, 2, batch_dims=[batch_dims] * 2, **TOLERANCES)



"""Unit tests of nonlinearity primitives."""

import pytest
import torch

from lgatr.primitives import gated_relu, gated_sigmoid, gated_silu
from lgatr.primitives.nonlinearities import gated_gelu
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("fn", [gated_relu, gated_gelu, gated_sigmoid, gated_silu])
@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_gated_nonlin_equivariance(fn, batch_dims):
    """Tests an identity map for equivariance (testing the test)."""
    gates = torch.randn(*batch_dims, 1)
    check_pin_equivariance(fn, 1, fn_kwargs=dict(gates=gates), batch_dims=batch_dims, **TOLERANCES)



import pytest
import torch

from lgatr.interface import embed_scalar, extract_scalar
from tests.helpers import BATCH_DIMS, TOLERANCES


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_embed_scalar(batch_dims):
    """Tests that embed_scalar() embeds scalars into multivectors correctly."""

    scalar = torch.randn(*batch_dims, 1)
    mv = embed_scalar(scalar)

    # Check that scalar part of MV contains the data we wanted to embed
    mv_scalar = mv[..., [0]]
    torch.testing.assert_close(mv_scalar, scalar, **TOLERANCES)

    # Check that other components of MV are empty
    mv_other = mv[..., 1:]
    torch.testing.assert_close(mv_other, torch.zeros_like(mv_other), **TOLERANCES)

    # Check that we can extract the scalar back from the multivector
    other_scalar = extract_scalar(embed_scalar(scalar))
    torch.testing.assert_close(scalar, other_scalar, **TOLERANCES)



from .axialvector import embed_axialvector, extract_axialvector
from .pseudoscalar import embed_pseudoscalar, extract_pseudoscalar
from .scalar import embed_scalar, extract_scalar
from .spurions import get_num_spurions, get_spurions
from .vector import embed_vector, extract_vector



"""Embedding and extracting axial vectors into multivectors."""

import torch


def embed_axialvector(axialvector: torch.Tensor) -> torch.Tensor:
    """Embeds axial vectors in multivectors.

    Parameters
    ----------
    axialvector : torch.Tensor
        Axial vector with shape (..., 4)

    Returns
    -------
    multivector : torch.Tensor
        Embedding into multivector with shape (..., 16).
    """

    # Create multivector tensor with same batch shape, same device, same dtype as input
    batch_shape = axialvector.shape[:-1]
    multivector = torch.zeros(*batch_shape, 16, dtype=axialvector.dtype, device=axialvector.device)

    # Embedding into Lorentz vectors
    multivector[..., 11:15] = axialvector.flip(-1)

    return multivector


def extract_axialvector(multivector: torch.Tensor) -> torch.Tensor:
    """Given a multivector, extract a axial vector.

    Parameters
    ----------
    multivector : torch.Tensor
        Multivector with shape (..., 16).

    Returns
    -------
    axialvector : torch.Tensor
        Axial vector with shape (..., 4)
    """

    axialvector = multivector[..., 11:15].flip(-1)

    return axialvector



"""Unit tests of bilinear primitives."""

import pytest

from lgatr.primitives import (
    abs_squared_norm,
    inner_product,
)
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_invariance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_inner_product_invariance(batch_dims):
    """Tests the innner_product() primitive for equivariance."""
    check_pin_invariance(inner_product, 2, batch_dims=[batch_dims] * 2, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_abs_squared_norm_invariance(batch_dims):
    """Tests the abs_squared_norm() primitive for equivariance."""
    check_pin_invariance(abs_squared_norm, 1, batch_dims=batch_dims, **TOLERANCES)



"""Equivariant attention."""

import torch
from einops import rearrange
from torch import Tensor

from .attention_backends import get_attention_backend
from .invariants import _load_inner_product_factors


def sdp_attention(
    q_mv: Tensor,
    k_mv: Tensor,
    v_mv: Tensor,
    q_s: Tensor,
    k_s: Tensor,
    v_s: Tensor,
    **attn_kwargs,
) -> tuple[Tensor, Tensor]:
    """Equivariant geometric attention based on scaled dot products.

    Expects both multivector and scalar queries, keys, and values as inputs.
    Then this function computes multivector and scalar outputs in the following way:

    .. code-block::

        attn_weights[..., i, j] = softmax_j[
            ga_inner_product(q_mv[..., i, :, :], k_mv[..., j, :, :])
            + euclidean_inner_product(q_s[..., i, :], k_s[..., j, :])
        ]
        out_mv[..., i, c, :] = sum_j attn_weights[..., i, j] v_mv[..., j, c, :] / norm
        out_s[..., i, c] = sum_j attn_weights[..., i, j] v_s[..., j, c] / norm

    Parameters
    ----------
    q_mv : torch.Tensor
        Multivector queries with shape (..., items_out, mv_channels, 16)
    k_mv : torch.Tensor
        Multivector keys with shape (..., items_out, mv_channels, 16)
    v_mv : torch.Tensor
        Multivector values with shape (..., items_out, mv_channels, 16)
    q_s : torch.Tensor
        Scalar queries with shape (..., items_out, s_channels)
    k_s : torch.Tensor
        Scalar keys with shape (..., items_out, s_channels)
    v_s : torch.Tensor
        Scalar values with shape (..., items_out, s_channels)
    **attn_kwargs
        Optional keyword arguments passed to attention.

    Returns
    -------
    outputs_mv : torch.Tensor
        Multivector result with shape (..., items_out, mv_channels, 16)
    outputs_s : torch.Tensor
        Scalar result with shape (..., items_out, s_channels)
    """

    # Construct queries and keys by concatenating relevant MV components and aux scalars
    q = torch.cat(
        [
            rearrange(
                q_mv * _load_inner_product_factors(device=q_mv.device, dtype=q_mv.dtype),
                "... c x -> ... (c x)",
            ),
            q_s,
        ],
        -1,
    )
    k = torch.cat([rearrange(k_mv, "... c x -> ... (c x)"), k_s], -1)

    num_channels_out = v_mv.shape[-2]
    v = torch.cat([rearrange(v_mv, "... c x -> ... (c x)"), v_s], -1)

    v_out = scaled_dot_product_attention(q, k, v, **attn_kwargs)

    v_out_mv = rearrange(v_out[..., : num_channels_out * 16], "... (c x) -> ...  c x", x=16)
    v_out_s = v_out[..., num_channels_out * 16 :]

    return v_out_mv, v_out_s


def scaled_dot_product_attention(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    **attn_kwargs,
) -> Tensor:
    """Execute scaled dot-product attention.
    The attention backend is determined dynamically
    based on the ``attn_kwargs`` provided.

    Parameters
    ----------
    query : torch.Tensor
        Tensor of shape (..., items_out, channels)
    key : torch.Tensor
        Tensor of shape (..., items_in, channels)
    value : torch.Tensor
        Tensor of shape (..., items_in, channels)
    **attn_kwargs
        Optional keyword arguments passed to attention.

    Returns
    -------
    torch.Tensor
        Tensor of shape (..., head, item_out, channels)
    """
    attention_backend = get_attention_backend(**attn_kwargs)
    return attention_backend(query, key, value, **attn_kwargs)



"""Equivariant conditional transformer for vector and scalar data."""

import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from ..primitives.attention import scaled_dot_product_attention
from .lgatr_slim import (
    MLP,
    Dropout,
    Linear,
    RMSNorm,
    SelfAttention,
)


class CrossAttention(nn.Module):
    """Cross-attention module for Lorentz vectors and scalar features."""

    def __init__(
        self,
        q_v_channels: int,
        kv_v_channels: int,
        q_s_channels: int,
        kv_s_channels: int,
        num_heads: int,
        attn_ratio: int = 1,
        dropout_prob: float | None = None,
    ):
        super().__init__()
        self.hidden_v_channels = max(attn_ratio * q_v_channels // num_heads, 1)
        self.hidden_s_channels = max(attn_ratio * q_s_channels // num_heads, 4)
        self.num_heads = num_heads

        metric = torch.tensor([1.0, -1.0, -1.0, -1.0])
        self.register_buffer("metric", metric)

        self.linear_in_q = Linear(
            in_v_channels=q_v_channels,
            out_v_channels=self.hidden_v_channels * self.num_heads,
            in_s_channels=q_s_channels,
            out_s_channels=self.hidden_s_channels * self.num_heads,
            initialization="small",
        )
        self.linear_in_kv = Linear(
            in_v_channels=kv_v_channels,
            out_v_channels=2 * self.hidden_v_channels * self.num_heads,
            in_s_channels=kv_s_channels,
            out_s_channels=2 * self.hidden_s_channels * self.num_heads,
            initialization="small",
        )
        self.linear_out = Linear(
            in_v_channels=self.hidden_v_channels * self.num_heads,
            out_v_channels=q_v_channels,
            in_s_channels=self.hidden_s_channels * self.num_heads,
            out_s_channels=q_s_channels,
            initialization="small",
        )

        self.norm = RMSNorm()
        if dropout_prob is not None:
            self.dropout = Dropout(dropout_prob)
        else:
            self.dropout = None

    def _pre_reshape(self, q_v, kv_v, q_s, kv_s):
        kv_v = (
            kv_v.unflatten(-2, (2, self.hidden_v_channels, self.num_heads))
            .movedim(-4, 0)
            .movedim(-2, -4)
        )  # (2, *B, H, N, Cv, 4)
        kv_s = (
            kv_s.unflatten(-1, (2, self.hidden_s_channels, self.num_heads))
            .movedim(-3, 0)
            .movedim(-1, -3)
        )  # (2, *B, H, N, Cs)
        q_v = q_v.unflatten(-2, (self.hidden_v_channels, self.num_heads)).movedim(
            -2, -4
        )  # (*B, H, Nc, Cv, 4)
        q_s = q_s.unflatten(-1, (self.hidden_s_channels, self.num_heads)).movedim(
            -1, -3
        )  # (*B, H, Nc, Cs)

        # normalize for stability (important)
        q_v, q_s = self.norm(q_v, q_s)
        kv_v, kv_s = self.norm(kv_v, kv_s)

        k_v, v_v = kv_v.unbind(0)
        k_s, v_s = kv_s.unbind(0)

        q_v_mod = q_v * self.metric
        q = torch.cat([q_v_mod.flatten(start_dim=-2), q_s], dim=-1)
        k = torch.cat([k_v.flatten(start_dim=-2), k_s], dim=-1)
        v = torch.cat([v_v.flatten(start_dim=-2), v_s], dim=-1)
        return q, k, v

    def _post_reshape(self, out):
        h_v = out[..., : self.hidden_v_channels * 4].reshape(
            *out.shape[:-1], self.hidden_v_channels, 4
        )
        h_s = out[..., self.hidden_v_channels * 4 :]

        h_v = h_v.movedim(-3, -4).flatten(-3, -2)
        h_s = h_s.movedim(-2, -3).flatten(-2, -1)
        return h_v, h_s

    def forward(self, vectors, vectors_condition, scalars, scalars_condition, **attn_kwargs):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.
        **attn_kwargs : dict
            Additional keyword arguments for the attention function.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        q_v, q_s = self.linear_in_q(vectors, scalars)
        kv_v, kv_s = self.linear_in_kv(vectors_condition, scalars_condition)

        q, k, v = self._pre_reshape(q_v, kv_v, q_s, kv_s)
        out = scaled_dot_product_attention(q, k, v, **attn_kwargs)
        h_v, h_s = self._post_reshape(out)

        out_v, out_s = self.linear_out(h_v, h_s)

        if self.dropout is not None:
            out_v, out_s = self.dropout(out_v, out_s)
        return out_v, out_s


class ConditionalLGATrSlimBlock(nn.Module):
    """A single block of the conditional L-GATr-slim,
    consisting of self-attention, cross-attention and MLP layers, pre-norm and residual connections."""

    def __init__(
        self,
        v_channels: int,
        condition_v_channels: int,
        s_channels: int,
        condition_s_channels: int,
        num_heads: int,
        nonlinearity: str = "gelu",
        mlp_ratio: int = 2,
        attn_ratio: int = 1,
        num_layers_mlp: int = 2,
        dropout_prob: float | None = None,
    ):
        super().__init__()

        self.norm = RMSNorm()

        self.selfattention = SelfAttention(
            v_channels=v_channels,
            s_channels=s_channels,
            num_heads=num_heads,
            attn_ratio=attn_ratio,
            dropout_prob=dropout_prob,
        )
        self.crossattention = CrossAttention(
            q_v_channels=v_channels,
            kv_v_channels=condition_v_channels,
            q_s_channels=s_channels,
            kv_s_channels=condition_s_channels,
            num_heads=num_heads,
            attn_ratio=attn_ratio,
            dropout_prob=dropout_prob,
        )

        self.mlp = MLP(
            v_channels=v_channels,
            s_channels=s_channels,
            nonlinearity=nonlinearity,
            mlp_ratio=mlp_ratio,
            num_layers=num_layers_mlp,
            dropout_prob=dropout_prob,
        )

    def forward(
        self,
        vectors,
        vectors_condition,
        scalars,
        scalars_condition,
        attn_kwargs=None,
        crossattn_kwargs=None,
    ):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        vectors_condition : torch.Tensor
            A tensor of shape (..., condition_v_channels, 4) representing a Lorentz vector condition included in cross-attention.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.
        scalars_condition : torch.Tensor
            A tensor of shape (..., condition_s_channels) representing a scalar condition included in cross-attention.
        **attn_kwargs : dict
            Additional keyword arguments for the attention function.
        **crossattn_kwargs : dict
            Additional keyword arguments for the cross-attention function.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        attn_kwargs = attn_kwargs if attn_kwargs is not None else {}
        crossattn_kwargs = crossattn_kwargs if crossattn_kwargs is not None else {}

        # self-attention block
        h_v, h_s = self.norm(vectors, scalars)
        h_v, h_s = self.selfattention(
            h_v,
            h_s,
            **attn_kwargs,
        )
        outputs_v = vectors + h_v
        outputs_s = scalars + h_s

        # cross-attention block
        h_v, h_s = self.norm(outputs_v, outputs_s)
        h_v, h_s = self.crossattention(
            h_v,
            vectors_condition,
            h_s,
            scalars_condition,
            **crossattn_kwargs,
        )
        outputs_v = outputs_v + h_v
        outputs_s = outputs_s + h_s

        # MLP block
        h_v, h_s = self.norm(outputs_v, outputs_s)
        h_v, h_s = self.mlp(h_v, h_s)
        outputs_v = outputs_v + h_v
        outputs_s = outputs_s + h_s

        return outputs_v, outputs_s


class ConditionalLGATrSlim(nn.Module):
    """Conditional L-GATr-slim network."""

    def __init__(
        self,
        in_v_channels: int,
        condition_v_channels: int,
        out_v_channels: int,
        hidden_v_channels: int,
        in_s_channels: int,
        condition_s_channels: int,
        out_s_channels: int,
        hidden_s_channels: int,
        num_blocks: int,
        num_heads: int,
        nonlinearity: str = "gelu",
        mlp_ratio: int = 2,
        attn_ratio: int = 1,
        num_layers_mlp: int = 2,
        dropout_prob: float | None = None,
        checkpoint_blocks: bool = False,
        compile: bool = False,
    ):
        """
        Parameters
        ----------
        in_v_channels : int
            Number of input vector channels.
        condition_v_channels : int
            Number of conditional vector channels.
        out_v_channels : int
            Number of output vector channels.
        hidden_v_channels : int
            Number of hidden vector channels.
        in_s_channels : int
            Number of input scalar channels.
        condition_s_channels : int
            Number of conditional scalar channels.
        out_s_channels : int
            Number of output scalar channels.
        hidden_s_channels : int
            Number of hidden scalar channels.
        num_blocks : int
            Number of Lorentz Transformer blocks.
        num_heads : int
            Number of attention heads.
        nonlinearity : str, optional
            Nonlinearity type for MLP layers, by default "gelu".
        mlp_ratio : int, optional
            Expansion ratio for MLP hidden layers, by default 2.
        attn_ratio : int, optional
            Expansion ratio for attention hidden layers, by default 1.
        num_layers_mlp : int, optional
            Number of layers in MLP, by default 2.
        dropout_prob : float | None, optional
            Dropout probability, by default None.
        checkpoint_blocks : bool, optional
            Whether to use gradient checkpointing for blocks, by default False.
        compile : bool, optional
            Whether to compile the model with torch.compile, by default False.
        """
        super().__init__()

        self.linear_in = Linear(
            in_v_channels=in_v_channels,
            in_s_channels=in_s_channels,
            out_v_channels=hidden_v_channels,
            out_s_channels=hidden_s_channels,
        )

        self.blocks = nn.ModuleList(
            [
                ConditionalLGATrSlimBlock(
                    v_channels=hidden_v_channels,
                    s_channels=hidden_s_channels,
                    condition_v_channels=condition_v_channels,
                    condition_s_channels=condition_s_channels,
                    num_heads=num_heads,
                    nonlinearity=nonlinearity,
                    mlp_ratio=mlp_ratio,
                    attn_ratio=attn_ratio,
                    num_layers_mlp=num_layers_mlp,
                    dropout_prob=dropout_prob,
                )
                for _ in range(num_blocks)
            ]
        )

        self.linear_out = Linear(
            in_v_channels=hidden_v_channels,
            in_s_channels=hidden_s_channels,
            out_v_channels=out_v_channels,
            out_s_channels=out_s_channels,
        )
        self._checkpoint_blocks = checkpoint_blocks

        self.compile = compile
        if compile:
            # ugly hack to make torch.compile convenient for users
            # the clean solution is model = torch.compile(model, **kwargs) outside of the constructor
            # note that we need fullgraph=False because of the torch.compiler.disable for attention
            self.__class__ = torch.compile(self.__class__, dynamic=True, mode="default")

    def forward(
        self,
        vectors,
        vectors_condition,
        scalars,
        scalars_condition,
        attn_kwargs=None,
        crossattn_kwargs=None,
    ):
        """
        Parameters
        ----------
        vectors : torch.Tensor
            A tensor of shape (..., v_channels, 4) representing Lorentz vectors.
        vectors_condition : torch.Tensor
            A tensor of shape (..., v_channels_condition, 4) representing a Lorentz vector condition included in cross-attention.
        scalars : torch.Tensor
            A tensor of shape (..., s_channels) representing scalar features.
        scalars_condition : torch.Tensor
            A tensor of shape (..., s_channels_condition) representing a scalar condition included in cross-attention.
        attn_kwargs : dict
            Additional keyword arguments for the self-attention function.
        crossattn_kwargs : dict
            Additional keyword arguments for the cross-attention function.

        Returns
        -------
        torch.Tensor, torch.Tensor
            Tensors of the same shape as input representing the normalized vectors and scalars.
        """
        attn_kwargs = attn_kwargs if attn_kwargs is not None else {}
        crossattn_kwargs = crossattn_kwargs if crossattn_kwargs is not None else {}

        h_v, h_s = self.linear_in(vectors, scalars)

        for block in self.blocks:
            if self._checkpoint_blocks:
                h_v, h_s = checkpoint(
                    block,
                    vectors=h_v,
                    scalars=h_s,
                    vectors_condition=vectors_condition,
                    scalars_condition=scalars_condition,
                    use_reentrant=False,
                    **attn_kwargs,
                )
            else:
                h_v, h_s = block(
                    vectors=h_v,
                    scalars=h_s,
                    vectors_condition=vectors_condition,
                    scalars_condition=scalars_condition,
                    **attn_kwargs,
                )

        outputs_v, outputs_s = self.linear_out(h_v, h_s)
        return outputs_v, outputs_s



import pytest
import torch

from lgatr.layers import GeoMLP
from lgatr.layers.mlp.config import MLPConfig
from lgatr.primitives.config import gatr_config
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_equivariance

_CHANNELS = [((5), (12)), ((4), (10)), ((4), None)]


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("activation", ["gelu"])
@pytest.mark.parametrize("mv_channels,s_channels", _CHANNELS)
@pytest.mark.parametrize("use_geometric_product", [True, False])
def test_geo_mlp_shape(batch_dims, mv_channels, s_channels, activation, use_geometric_product):
    """Tests the output shape of GeoMLP()."""
    gatr_config.use_geometric_product = use_geometric_product

    inputs = torch.randn(*batch_dims, mv_channels, 16)
    scalars = None if s_channels is None else torch.randn(*batch_dims, s_channels)

    try:
        net = GeoMLP(
            MLPConfig(mv_channels=mv_channels, s_channels=s_channels, activation=activation)
        )
    except NotImplementedError:
        return  # "GeoMLP not implemented for this configuration"
    outputs, outputs_scalars = net(inputs, scalars=scalars)

    assert outputs.shape == (*batch_dims, mv_channels, 16)
    if s_channels is not None:
        assert outputs_scalars.shape == (*batch_dims, s_channels)


@pytest.mark.parametrize("batch_dims", [[100]])
@pytest.mark.parametrize("activation", ["gelu"])
@pytest.mark.parametrize("mv_channels,s_channels", _CHANNELS)
def test_geo_mlp_equivariance(batch_dims, mv_channels, s_channels, activation):
    """Tests GeoMLP() for Pin equivariance."""
    net = GeoMLP(MLPConfig(mv_channels=mv_channels, s_channels=s_channels, activation=activation))
    data_dims = tuple(list(batch_dims) + [mv_channels])
    scalars = None if s_channels is None else torch.randn(*batch_dims, s_channels)

    # Because of the fixed reference MV, we only test Spin equivariance
    check_pin_equivariance(
        net,
        1,
        batch_dims=data_dims,
        fn_kwargs=dict(scalars=scalars),
        spin=True,
        **TOLERANCES,
    )



"""Geometric product on multivectors."""

import torch
from torch import nn

from ...primitives import geometric_product
from ...primitives.config import gatr_config
from ..layer_norm import EquiLayerNorm
from ..linear import EquiLinear


class GeometricBilinear(nn.Module):
    """Geometric product on multivectors.

    Pin-equivariant map between multivector tensors that constructs new geometric features via
    geometric products.

    Parameters
    ----------
    in_mv_channels : int
        Input multivector channels
    out_mv_channels : int
        Output multivector channels
    hidden_mv_channels : int or None
        Hidden multivector channels. If None, uses out_mv_channels.
    in_s_channels : int or None
        Input scalar channels. If None, no scalars are expected nor returned.
    out_s_channels : int or None
        Output scalar channels. If None, no scalars are expected nor returned.
    """

    def __init__(
        self,
        in_mv_channels: int,
        out_mv_channels: int,
        hidden_mv_channels: int | None = None,
        in_s_channels: int | None = None,
        out_s_channels: int | None = None,
    ) -> None:
        super().__init__()

        # Default options
        if hidden_mv_channels is None:
            hidden_mv_channels = out_mv_channels

        # Linear projections for GP
        self.linear_left = EquiLinear(
            in_mv_channels,
            hidden_mv_channels,
            in_s_channels=in_s_channels,
            out_s_channels=None,
        )
        self.linear_right = EquiLinear(
            in_mv_channels,
            hidden_mv_channels,
            in_s_channels=in_s_channels,
            out_s_channels=None,
            initialization="almost_unit_scalar",
        )

        # Output linear projection
        self.linear_out = EquiLinear(
            hidden_mv_channels, out_mv_channels, in_s_channels, out_s_channels
        )
        self.norm = EquiLayerNorm()

    def forward(
        self,
        multivectors: torch.Tensor,
        scalars: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., in_mv_channels, 16)
        scalars : torch.Tensor with shape (..., in_s_channels)
            Input scalars

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., out_mv_channels, 16)
        output_s : None or torch.Tensor
            Output scalars with shape (..., out_s_channels).
        """

        # GP
        left, _ = self.linear_left(multivectors, scalars=scalars)
        right, _ = self.linear_right(multivectors, scalars=scalars)
        gp_outputs = geometric_product(left, right)
        if not gatr_config.use_bivector:
            gp_outputs[..., 5:11] = 0.0

        # Output linear
        outputs_mv, outputs_s = self.linear_out(gp_outputs, scalars=scalars)

        outputs_mv, outputs_s = self.norm(outputs_mv, outputs_s)
        return outputs_mv, outputs_s



"""Equivariant dropout layer."""

import torch
from torch import nn

from ..primitives import grade_dropout


class GradeDropout(nn.Module):
    """Dropout on multivectors.

    Parameters
    ----------
    p : float
        Dropout probability.
    """

    def __init__(self, p: float = 0.0):
        super().__init__()
        self._dropout_prob = p

    def forward(
        self, multivectors: torch.Tensor, scalars: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass. Applies dropout.

        Parameters
        ----------
        multivectors : torch.Tensor
            Multivector inputs  with shape (..., 16).
        scalars : torch.Tensor
            Scalar inputs with shape (...).

        Returns
        -------
        outputs_mv : torch.Tensor
            Multivector inputs with dropout applied, shape (..., 16).
        output_scalars : torch.Tensor
            Scalar inputs with dropout applied, shape (...).
        """

        out_mv = grade_dropout(multivectors, p=self._dropout_prob, training=self.training)
        out_s = torch.nn.functional.dropout(scalars, p=self._dropout_prob, training=self.training)

        return out_mv, out_s



"""Gated nonlinearities on multivectors."""

import torch


def gated_relu(x: torch.Tensor, gates: torch.Tensor) -> torch.Tensor:
    """Pin-equivariant gated ReLU nonlinearity.

    Given multivector input x and scalar input gates (with matching batch dimensions), computes
    ``ReLU(gates) * x``.

    Parameters
    ----------
    x : torch.Tensor
        Multivector input with shape (..., 16)
    gates : torch.Tensor
        Pin-invariant gates with shape (..., 1).

    Returns
    -------
    outputs : torch.Tensor
        Computes ReLU(gates) * x, with broadcasting along the last dimension.
        Result has shape (..., 16)
    """

    weights = torch.nn.functional.relu(gates)
    outputs = weights * x
    return outputs


def gated_sigmoid(x: torch.Tensor, gates: torch.Tensor):
    """Pin-equivariant gated sigmoid nonlinearity.

    Given multivector input x and scalar input gates (with matching batch dimensions), computes
    ``sigmoid(gates) * x``.

    Parameters
    ----------
    x : torch.Tensor
        Multivector input with shape (..., 16)
    gates : torch.Tensor
        Pin-invariant gates with shape (..., 1).

    Returns
    -------
    outputs : torch.Tensor
        Computes Sigmoid(gates) * x, with broadcasting along the last dimension.
        Result has shape (..., 16)
    """

    weights = torch.nn.functional.sigmoid(gates)
    outputs = weights * x
    return outputs


def gated_gelu(x: torch.Tensor, gates: torch.Tensor) -> torch.Tensor:
    """Pin-equivariant gated GeLU nonlinearity without division.

    Given multivector input x and scalar input gates (with matching batch dimensions), computes
    ``GeLU(gates) * x``.

    References
    ----------
    Dan Hendrycks, Kevin Gimpel, "Gaussian Error Linear Units (GELUs)", arXiv:1606.08415

    Parameters
    ----------
    x : torch.Tensor
        Multivector input with shape (..., 16)
    gates : torch.Tensor
        Pin-invariant gates with shape (..., 1).

    Returns
    -------
    outputs : torch.Tensor
        Computes GeLU(gates) * x, with broadcasting along the last dimension.
        Result has shape (..., 16)
    """

    weights = torch.nn.functional.gelu(gates, approximate="tanh")
    outputs = weights * x
    return outputs


def gated_silu(x: torch.Tensor, gates: torch.Tensor) -> torch.Tensor:
    """Pin-equivariant gated SiLU or Swish nonlinearity without division.

    Given multivector input x and scalar input gates (with matching batch dimensions), computes
    ``Swish(gates) * x``.

    References
    ----------
    Stefan Elfwing, Eiji Uchibe, Kenji Doya, "Sigmoid-Weighted Linear Units for Neural Network
    Function Approximation in Reinforcement Learning", arXiv:1702.03118

    Parameters
    ----------
    x : torch.Tensor
        Multivector input with shape (..., 16)
    gates : torch.Tensor
        Pin-invariant gates with shape (..., 1).

    Returns
    -------
    outputs : torch.Tensor
        Computes Swish(gates) * x, with broadcasting along the last dimension.
        Result has shape (..., 16)
    """

    weights = torch.nn.functional.silu(gates)
    outputs = weights * x
    return outputs



import importlib.util

import pytest
import torch
from torch.nn.functional import scaled_dot_product_attention as torch_sdpa

from lgatr.primitives.attention_backends import get_attention_backend
from tests.helpers.constants import STRICT_TOLERANCES as TOLERANCES

SHAPES = [
    (32, 8, 5, 32),
    (9, 3, 7, 13),
]


_xformers_available = importlib.util.find_spec("xformers") is not None
_flash_available = importlib.util.find_spec("flash-attn") is not None
_torch_version = torch.__version__.split("+")[0]
_flex_available = tuple(int(x) for x in _torch_version.split(".")[:2]) >= (2, 7)
_varlen_available = tuple(int(x) for x in _torch_version.split(".")[:2]) >= (2, 10)


def _random_qkv(shape, dtype=torch.float32, device=None):
    if device is None:
        device = torch.device("cpu")

    q = torch.randn(*shape, dtype=dtype, device=device)
    k = torch.randn(*shape, dtype=dtype, device=device)
    v = torch.randn(*shape, dtype=dtype, device=device)
    return q, k, v


def _sparsify_shape(shape_dense, device=None):
    if device is None:
        device = torch.device("cpu")

    shape_sparse = (1, shape_dense[1], shape_dense[0] * shape_dense[2], shape_dense[3])
    cu_seq = [shape_dense[2] * i for i in range(shape_dense[0] + 1)]
    cu_seq = torch.tensor(cu_seq, dtype=torch.int32, device=device)
    max_seq = shape_dense[2]
    return shape_sparse, cu_seq, max_seq


@pytest.mark.parametrize("shape", SHAPES)
def test_default_backend_selection(shape):
    backend_fn = get_attention_backend()
    assert backend_fn is torch_sdpa

    qkv = _random_qkv(shape)
    out = backend_fn(*qkv)
    assert out.shape == shape


@pytest.mark.skipif(not _xformers_available, reason="xformers not installed")
@pytest.mark.parametrize("shape", SHAPES)
def test_xformers_backend_selection(shape):
    from lgatr.primitives.attention_backends.xformers import attention

    # check that backend exists
    backend_fn = get_attention_backend(backend="xformers")
    assert backend_fn is attention

    # check that result has correct shape
    qkv = _random_qkv(shape)
    out = backend_fn(*qkv)
    assert out.shape == shape

    # check agreement with default attention
    default_backend_fn = get_attention_backend()
    out_default = default_backend_fn(*qkv)
    torch.testing.assert_close(out, out_default, **TOLERANCES)


@pytest.mark.skipif(not _flex_available, reason="flex requires torch>=2.7")
@pytest.mark.parametrize("shape", SHAPES)
def test_flex_backend_selection(shape):
    from lgatr.primitives.attention_backends.flex import attention

    # check that backend exists
    backend_fn = get_attention_backend(backend="flex")
    assert backend_fn is attention

    # check that result has correct shape
    qkv = _random_qkv(shape)
    out = backend_fn(*qkv)
    assert out.shape == shape

    # check agreement with default attention
    default_backend_fn = get_attention_backend()
    out_default = default_backend_fn(*qkv)
    torch.testing.assert_close(out, out_default, **TOLERANCES)


@pytest.mark.skipif(
    not _flash_available or not torch.cuda.is_available(),
    reason="flash requires flash-attn package and CUDA",
)
@pytest.mark.parametrize("shape", [SHAPES[0]])
def test_flash_backend_selection(shape):
    from lgatr.primitives.attention_backends.flash import attention

    # check that backend exists
    backend_fn = get_attention_backend(backend="flash")
    assert backend_fn is attention

    # check that result has correct shape
    device = torch.device("cuda")
    shape, cu_seq, max_seq = _sparsify_shape(shape, device=device)
    kwargs = {
        "cu_seqlens_q": cu_seq,
        "cu_seqlens_k": cu_seq,
        "max_seqlen_q": max_seq,
        "max_seqlen_k": max_seq,
    }
    qkv = _random_qkv(shape, device=device)
    out = backend_fn(*qkv, **kwargs)
    assert out.shape == shape


@pytest.mark.skipif(
    not _varlen_available or not torch.cuda.is_available(),
    reason="varlen requires torch>=2.10 and CUDA",
)
@pytest.mark.parametrize("shape", [SHAPES[0]])
def test_varlen_backend_selection(shape):
    from lgatr.primitives.attention_backends.varlen import attention

    # check that backend exists
    backend_fn = get_attention_backend(backend="varlen")
    assert backend_fn is attention

    # check that result has correct shape
    device = torch.device("cuda")
    shape, cu_seq, max_seq = _sparsify_shape(shape, device=device)
    kwargs = {"cu_seq_q": cu_seq, "cu_seq_k": cu_seq, "max_q": max_seq, "max_k": max_seq}
    qkv = _random_qkv(shape, device=device)
    out = backend_fn(*qkv, **kwargs)
    assert out.shape == shape



"""Multivector normalization."""

import torch

from ..utils.misc import minimum_autocast_precision
from .invariants import abs_squared_norm


@minimum_autocast_precision(torch.float32)
def equi_layer_norm(
    x: torch.Tensor, channel_dim: int = -2, gain: float = 1.0, epsilon: float = 0.01
) -> torch.Tensor:
    """Equivariant LayerNorm for multivectors.

    Rescales input such that ``mean_channels |inputs|^2 = 1``, where the norm is the GA norm and the
    mean goes over the channel dimensions.

    Using a factor ``gain > 1`` makes up for the fact that the GP norm overestimates the actual
    standard deviation of the input data.

    Parameters
    ----------
    x : torch.Tensor
        Multivectors with shape (..., 16).
    channel_dim : int
        Channel dimension index. Defaults to the second-last entry (last are the multivector components).
    gain : float
        Target output scale.
    epsilon : float
        Small numerical factor to avoid instabilities. By default, we use a reasonably large number
        to balance issues that arise from some multivector components not contributing to the norm.

    Returns
    -------
    outputs : torch.Tensor
        Normalized multivectors with shape (..., 16).
    """

    # Compute mean_channels |inputs|^2
    abs_squared_norms = abs_squared_norm(x)
    abs_squared_norms = torch.mean(abs_squared_norms, dim=channel_dim, keepdim=True)

    # Insure against low-norm tensors (which can arise even when `x.var(dim=-1)` is high b/c some
    # entries don't contribute to the inner product / GP norm!)
    abs_squared_norms = torch.clamp(abs_squared_norms, epsilon)

    # Rescale inputs
    outputs = gain * x * torch.rsqrt(abs_squared_norms)

    return outputs



from collections.abc import Callable
from functools import wraps
from itertools import chain
from typing import Any, Literal

import torch
from torch import Tensor


def minimum_autocast_precision(
    min_dtype: torch.dtype = torch.float32,
    output: Literal["low", "high"] | torch.dtype | None = None,
    which_args: list[int] | None = None,
    which_kwargs: list[str] | None = None,
):
    """Decorator that ensures input tensors are autocast to a minimum precision.
    Only has an effect in autocast-enabled regions. Otherwise, does not change the function.
    Only floating-point inputs are modified. Non-tensors, integer tensors, and boolean tensors are
    untouched.
    Note: AMP is turned on and off separately for CPU and CUDA. This decorator may fail in
    the case where both devices are used, with only one of them on AMP.

    Parameters
    ----------
    min_dtype : dtype
        Minimum dtype. Default: float32.
    output: None or "low" or "high" or dtype
        Specifies which dtypes the outputs should be cast to. Only floating-point Tensor outputs
        are affected. If None, the outputs are not modified. If "low", the lowest-precision input
        dtype is used. If "high", `min_dtype` or the highest-precision input dtype is used
        (whichever is higher).
    which_args : None or list of int
        If not None, specifies which positional arguments are to be modified. If None (the default),
        all positional arguments are modified (if they are Tensors and of a floating-point dtype).
    which_kwargs : bool
        If not None, specifies which keyword arguments are to be modified. If None (the default),
        all keyword arguments are modified (if they are Tensors and of a floating-point dtype).

    Returns
    -------
    decorator : Callable
        Decorator.
    """

    def decorator(func: Callable):
        """Decorator that casts input tensors to minimum precision."""

        def _cast_in(var: Any):
            """Casts a single input to at least 32-bit precision."""
            if not isinstance(var, Tensor):
                # We don't want to modify non-Tensors
                return var
            if not var.dtype.is_floating_point:
                # Integer / boolean tensors are also not touched
                return var
            if torch.finfo(var.dtype).bits >= torch.finfo(min_dtype).bits:
                return var
            return var.to(min_dtype)

        def _cast_out(var: Any, dtype: torch.dtype):
            """Casts a single output to desired precision."""
            if not isinstance(var, Tensor):
                # We don't want to modify non-Tensors
                return var
            if not var.dtype.is_floating_point:
                # Integer / boolean tensors are also not touched
                return var
            return var.to(dtype)

        @wraps(func)
        def decorated_func(*args: Any, **kwargs: Any):
            """Decorated func."""
            # Only change dtypes in autocast-enabled regions
            if not (torch.is_autocast_enabled() or torch.is_autocast_cpu_enabled()):
                # NB: torch.is_autocast_enabled() only checks for GPU autocast
                # See https://github.com/pytorch/pytorch/issues/110966
                return func(*args, **kwargs)
            # Cast inputs to at least 32 bit
            mod_args = [
                _cast_in(arg) for i, arg in enumerate(args) if which_args is None or i in which_args
            ]
            mod_kwargs = {
                key: _cast_in(val)
                for key, val in kwargs.items()
                if which_kwargs is None or key in which_kwargs
            }
            # Call function w/o autocast enabled
            with (
                torch.autocast(device_type="cuda", enabled=False),
                torch.autocast(device_type="cpu", enabled=False),
            ):
                outputs = func(*mod_args, **mod_kwargs)
            # Cast outputs to correct dtype
            if output is None:
                return outputs
            if output in ["low", "high"]:
                in_dtypes = [
                    arg.dtype
                    for arg in chain(args, kwargs.values())
                    if isinstance(arg, Tensor) and arg.dtype.is_floating_point
                ]
                assert len(in_dtypes)
                if output == "low":
                    out_dtype = min([min_dtype] + in_dtypes, key=lambda dt: torch.finfo(dt).bits)
                else:
                    out_dtype = max(in_dtypes, key=lambda dt: torch.finfo(dt).bits)
            else:
                out_dtype = output
            if isinstance(outputs, tuple):
                return (_cast_out(val, out_dtype) for val in outputs)
            else:
                return _cast_out(outputs, out_dtype)

        return decorated_func

    return decorator



import pytest
import torch

from lgatr.interface.spurions import get_num_spurions, get_spurions


@pytest.mark.parametrize(
    "beam_spurion, add_time_spurion, beam_mirror, expected_num",
    [
        # beam_spurion = "xyplane"
        ("xyplane", True, True, 2),  # +1 for xyplane +1 for time
        ("xyplane", False, True, 1),
        ("xyplane", True, False, 2),
        ("xyplane", False, False, 1),
        # beam_spurion = "lightlike"
        ("lightlike", True, True, 3),  # +2 for mirror +1 for time
        ("lightlike", True, False, 2),
        ("lightlike", False, True, 2),
        ("lightlike", False, False, 1),
        # beam_spurion = "spacelike"
        ("spacelike", True, True, 3),
        ("spacelike", False, True, 2),
        ("spacelike", True, False, 2),
        ("spacelike", False, False, 1),
        # beam_spurion = "timelike"
        ("timelike", True, True, 3),
        ("timelike", False, True, 2),
        ("timelike", True, False, 2),
        ("timelike", False, False, 1),
        # beam_spurion = None
        (None, True, True, 1),  # +0 for beam +1 for time
        (None, False, True, 0),
        (None, True, False, 1),
        (None, False, False, 0),
    ],
)
def test_get_num_spurions(beam_spurion, add_time_spurion, beam_mirror, expected_num):
    """Check that get_num_spurions returns the correct integer for each config."""
    result = get_num_spurions(
        beam_spurion=beam_spurion,
        add_time_spurion=add_time_spurion,
        beam_mirror=beam_mirror,
    )
    assert result == expected_num


@pytest.mark.parametrize(
    "beam_spurion, add_time_spurion, beam_mirror",
    [
        ("xyplane", True, True),
        ("xyplane", False, False),
        ("lightlike", True, True),
        ("lightlike", False, False),
        ("spacelike", True, True),
        ("spacelike", False, False),
        ("timelike", True, True),
        ("timelike", False, False),
        (None, True, True),
        (None, False, False),
    ],
)
def test_get_spurions(beam_spurion, add_time_spurion, beam_mirror):
    """
    Verify that get_spurions returns a tensor of shape (n_spurions, 16),
    where n_spurions = get_num_spurions(...).
    """
    expected_num = get_num_spurions(
        beam_spurion=beam_spurion,
        add_time_spurion=add_time_spurion,
        beam_mirror=beam_mirror,
    )
    result = get_spurions(
        beam_spurion=beam_spurion,
        add_time_spurion=add_time_spurion,
        beam_mirror=beam_mirror,
        device="cpu",
        dtype=torch.float32,
    )

    # Check type
    assert isinstance(result, torch.Tensor), "Output should be a torch.Tensor."

    # Check shape
    assert result.shape == (expected_num, 16)



"""Tools to include reference multivectors ('spurions') for symmetry breaking."""

import torch

from .vector import embed_vector


def get_num_spurions(
    beam_spurion="xyplane",
    add_time_spurion=True,
    beam_mirror=True,
):
    """
    Compute how many reference multivectors/spurions a given configuration will have

    Parameters
    ----------
    beam_spurion: str
        Different options for adding a beam_reference
        Options: 'lightlike', 'spacelike', 'timelike', 'xyplane', None
    add_time_spurion: bool
        Whether to add the time direction as a reference to the network
    beam_mirror: bool
        If true, include (x, 0, 0, -1) in addition to (x, 0, 0, 1).
        Only relevant for beam_reference in ['lightlike', 'spacelike', 'timelike']
        This is not required for ``beam_spurion=xyplane``, as the xy-plane is symmetric

    Returns
    -------
    num_spurions: int
        Number of spurions
    """
    assert beam_spurion in ["xyplane", "lightlike", "spacelike", "timelike", None]

    num_spurions = 0
    if beam_spurion in ["lightlike", "spacelike", "timelike"]:
        num_spurions += 2 if beam_mirror else 1
    elif beam_spurion == "xyplane":
        num_spurions += 1
    if add_time_spurion:
        num_spurions += 1
    return num_spurions


def get_spurions(
    beam_spurion="xyplane",
    add_time_spurion=True,
    beam_mirror=True,
    device="cpu",
    dtype=torch.float32,
):
    """
    Construct a list of reference multivectors/spurions for symmetry breaking

    Parameters
    ----------
    beam_spurion: str
        Different options for adding a beam_reference
        Options: 'lightlike', 'spacelike', 'timelike', 'xyplane', None
    add_time_spurion: bool
        Whether to add the time direction as a reference to the network
    beam_mirror: bool
        Whether we only want (x, 0, 0, 1) or both (x, 0, 0, +/- 1) for the beam
        If true, include (x, 0, 0, -1) in addition to (x, 0, 0, 1).
        Only relevant for beam_reference in ['lightlike', 'spacelike', 'timelike']
        This is not required for ``beam_spurion=xyplane``, as the xy-plane is symmetric
    device: torch.device
    dtype: torch.dtype

    Returns
    -------
    spurions: torch.tensor
        spurion embedded as multivector object with shape (num_spurions, 16)
    """
    assert beam_spurion in ["xyplane", "lightlike", "spacelike", "timelike", None]
    kwargs = {"device": device, "dtype": dtype}

    if beam_spurion in ["lightlike", "spacelike", "timelike"]:
        # add another 4-momentum
        if beam_spurion == "lightlike":
            beam = [1, 0, 0, 1]
        elif beam_spurion == "timelike":
            beam = [2**0.5, 0, 0, 1]
        elif beam_spurion == "spacelike":
            beam = [0, 0, 0, 1]
        beam = torch.tensor(beam, **kwargs).reshape(1, 4)
        beam = embed_vector(beam)
        if beam_mirror:
            beam2 = beam.clone()
            beam2[..., 4] = -1  # flip pz
            beam = torch.cat((beam, beam2), dim=0)

    elif beam_spurion == "xyplane":
        # add the x-y-plane, embedded as a bivector
        # convention for bivector components: [tx, ty, tz, xy, xz, yz]
        beam = torch.zeros(1, 16, **kwargs)
        beam[..., 8] = 1

    elif beam_spurion is None:
        beam = torch.empty(0, 16, **kwargs)

    if add_time_spurion:
        time = [1, 0, 0, 0]
        time = torch.tensor(time, **kwargs).unsqueeze(0)
        time = embed_vector(time)
    else:
        time = torch.empty(0, 16, **kwargs)

    spurions = torch.cat((beam, time), dim=-2)
    return spurions



import pytest
import torch

from lgatr.layers import LGATrBlock
from lgatr.layers.attention.config import SelfAttentionConfig
from lgatr.layers.mlp.config import MLPConfig
from tests.helpers import BATCH_DIMS, MILD_TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("num_items,mv_channels", [(8, 6)])
@pytest.mark.parametrize("num_heads", [4, 1])
@pytest.mark.parametrize("s_channels", [2, 6])
@pytest.mark.parametrize("dropout_prob", [None, 0.0, 0.3])
@pytest.mark.parametrize("multi_query_attention", [False, True])
def test_lgatr_block_shape(
    batch_dims,
    num_items,
    mv_channels,
    num_heads,
    s_channels,
    multi_query_attention,
    dropout_prob,
):
    """Tests the output shape of LGATrBlock."""
    inputs = torch.randn(*batch_dims, num_items, mv_channels, 16)
    scalars = None if s_channels is None else torch.randn(*batch_dims, num_items, s_channels)

    try:
        net = LGATrBlock(
            mv_channels,
            s_channels=s_channels,
            attention=SelfAttentionConfig(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            mlp=MLPConfig(),
            dropout_prob=dropout_prob,
        )
    except NotImplementedError:
        # Some features require scalar inputs, and failing without them is fine
        return

    outputs, output_scalars = net(inputs, scalars=scalars)

    assert outputs.shape == (*batch_dims, num_items, mv_channels, 16)
    if s_channels is not None:
        assert output_scalars.shape == (*batch_dims, num_items, s_channels)


@pytest.mark.parametrize("batch_dims", [(64,)])
@pytest.mark.parametrize("num_items,mv_channels", [(8, 6)])
@pytest.mark.parametrize("num_heads", [4, 1])
@pytest.mark.parametrize("s_channels", [2, 6])
@pytest.mark.parametrize("multi_query_attention", [False, True])
def test_lgatr_block_equivariance(
    batch_dims,
    num_items,
    mv_channels,
    num_heads,
    s_channels,
    multi_query_attention,
):
    """Tests LGATrBlock for equivariance."""
    try:
        net = LGATrBlock(
            mv_channels,
            s_channels=s_channels,
            attention=SelfAttentionConfig(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            mlp=MLPConfig(),
        )
    except NotImplementedError:
        # Some features require scalar inputs, and failing without them is fine
        return

    scalars = None if s_channels is None else torch.randn(*batch_dims, num_items, s_channels)
    data_dims = tuple(list(batch_dims) + [num_items, mv_channels])
    check_pin_equivariance(
        net, 1, batch_dims=data_dims, fn_kwargs=dict(scalars=scalars), **MILD_TOLERANCES
    )



"""L-GATr cross-attention."""

import torch
from einops import rearrange
from torch import nn

from ..dropout import GradeDropout
from ..linear import EquiLinear
from .attention import GeometricAttention
from .config import CrossAttentionConfig


class CrossAttention(nn.Module):
    """L-GATr cross-attention.

    Constructs queries, keys, and values, computes attention, and projects linearly to outputs.

    Parameters
    ----------
    config : CrossAttentionConfig
        Attention configuration.
    """

    def __init__(
        self,
        config: CrossAttentionConfig,
    ) -> None:
        super().__init__()

        if (
            config.additional_q_mv_channels > 0
            or config.additional_q_s_channels > 0
            or config.additional_k_mv_channels > 0
            or config.additional_k_s_channels > 0
        ):
            raise NotImplementedError("Cross attention is not implemented with additional channels")

        # Store settings
        self.config = config

        self.q_linear = EquiLinear(
            in_mv_channels=config.in_q_mv_channels,
            out_mv_channels=config.hidden_mv_channels * config.num_heads,
            in_s_channels=config.in_q_s_channels,
            out_s_channels=config.hidden_s_channels * config.num_heads,
        )
        self.kv_linear = EquiLinear(
            in_mv_channels=config.in_kv_mv_channels,
            out_mv_channels=2
            * config.hidden_mv_channels
            * (1 if config.multi_query else config.num_heads),
            in_s_channels=config.in_kv_s_channels,
            out_s_channels=2
            * config.hidden_s_channels
            * (1 if config.multi_query else config.num_heads),
        )

        # Output projection
        self.out_linear = EquiLinear(
            in_mv_channels=config.hidden_mv_channels * config.num_heads,
            out_mv_channels=config.out_mv_channels,
            in_s_channels=(
                None
                if config.in_kv_s_channels is None
                else config.hidden_s_channels * config.num_heads
            ),
            out_s_channels=config.out_s_channels,
            initialization=config.output_init,
        )

        # Attention
        self.attention = GeometricAttention(config)

        # Dropout
        self.dropout: nn.Module | None
        if config.dropout_prob is not None:
            self.dropout = GradeDropout(config.dropout_prob)
        else:
            self.dropout = None

        # HeadScaleMHA
        self.use_head_scale = config.head_scale
        if self.use_head_scale:
            self.head_scale = nn.Parameter(torch.ones(config.num_heads))

    def forward(
        self,
        multivectors_kv: torch.Tensor,
        multivectors_q: torch.Tensor,
        scalars_kv: torch.Tensor | None = None,
        scalars_q: torch.Tensor | None = None,
        **attn_kwargs,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute cross attention.

        The result is the following:

        .. code-block::

            # For each head
            queries = linear_channels(inputs_q)
            keys = linear_channels(inputs_kv)
            values = linear_channels(inputs_kv)
            hidden = attention_items(queries, keys, values, biases=biases)
            head_output = linear_channels(hidden)

            # Combine results
            output = concatenate_heads head_output

        Parameters
        ----------
        multivectors_kv : torch.Tensor
            Input multivectors for key and value with shape (..., items_kv, mv_channels, 16).
        multivectors_q : torch.Tensor
            Input multivectors for query with shape (..., items_q, mv_channels, 16).
        scalars_kv : None or torch.Tensor
            Optional input scalars for key and value with shape (..., items_kv, s_channels)
        scalars_q : None or torch.Tensor
            Optional input scalars for query with shape (..., items_q, s_channels)
        **attn_kwargs
            Optional keyword arguments passed to attention.

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., items_q, mv_channels, 16).
        output_scalars : torch.Tensor
            Output scalars with shape (..., items_q, s_channels).
        """
        q_mv, q_s = self.q_linear(
            multivectors_q, scalars_q
        )  # (..., num_items, hidden_channels, 16)
        kv_mv, kv_s = self.kv_linear(
            multivectors_kv, scalars_kv
        )  # (..., num_items, 2*hidden_channels, 16)
        k_mv, v_mv = torch.tensor_split(kv_mv, 2, dim=-2)
        k_s, v_s = torch.tensor_split(kv_s, 2, dim=-1)

        # Rearrange to (..., heads, items, channels, 16) shape
        q_mv = rearrange(
            q_mv,
            "... items (hidden_channels num_heads) x -> ... num_heads items hidden_channels x",
            num_heads=self.config.num_heads,
            hidden_channels=self.config.hidden_mv_channels,
        )
        if self.config.multi_query:
            k_mv = rearrange(k_mv, "... items hidden_channels x -> ... 1 items hidden_channels x")
            v_mv = rearrange(v_mv, "... items hidden_channels x -> ... 1 items hidden_channels x")
        else:
            k_mv = rearrange(
                k_mv,
                "... items (hidden_channels num_heads) x -> ... num_heads items hidden_channels x",
                num_heads=self.config.num_heads,
                hidden_channels=self.config.hidden_mv_channels,
            )
            v_mv = rearrange(
                v_mv,
                "... items (hidden_channels num_heads) x -> ... num_heads items hidden_channels x",
                num_heads=self.config.num_heads,
                hidden_channels=self.config.hidden_mv_channels,
            )

        # Same for scalars
        if q_s is not None:
            q_s = rearrange(
                q_s,
                "... items (hidden_channels num_heads) -> ... num_heads items hidden_channels",
                num_heads=self.config.num_heads,
                hidden_channels=self.config.hidden_s_channels,
            )
            if self.config.multi_query:
                k_s = rearrange(k_s, "... items hidden_channels -> ... 1 items hidden_channels")
                v_s = rearrange(v_s, "... items hidden_channels -> ... 1 items hidden_channels")
            else:
                k_s = rearrange(
                    k_s,
                    "... items (hidden_channels num_heads) -> ... num_heads items hidden_channels",
                    num_heads=self.config.num_heads,
                    hidden_channels=self.config.hidden_s_channels,
                )
                v_s = rearrange(
                    v_s,
                    "... items (hidden_channels num_heads) -> ... num_heads items hidden_channels",
                    num_heads=self.config.num_heads,
                    hidden_channels=self.config.hidden_s_channels,
                )
        else:
            q_s, k_s, v_s = None, None, None

        # Attention layer
        h_mv, h_s = self.attention(
            q_mv,
            k_mv,
            v_mv,
            q_s,
            k_s,
            v_s,
            **attn_kwargs,
        )
        if self.use_head_scale:
            h_mv = h_mv * self.head_scale.view(
                *[1] * len(h_mv.shape[:-5]), len(self.head_scale), 1, 1, 1
            )
            h_s = h_s * self.head_scale.view(*[1] * len(h_s.shape[:-4]), len(self.head_scale), 1, 1)

        h_mv = rearrange(
            h_mv,
            "... n_heads n_items hidden_channels x -> ... n_items (n_heads hidden_channels) x",
        )
        h_s = rearrange(
            h_s,
            "... n_heads n_items hidden_channels -> ... n_items (n_heads hidden_channels)",
        )

        # Transform linearly one more time
        outputs_mv, outputs_s = self.out_linear(h_mv, scalars=h_s)

        # Dropout
        if self.dropout is not None:
            outputs_mv, outputs_s = self.dropout(outputs_mv, outputs_s)

        return outputs_mv, outputs_s
